import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    snmpEngine = SnmpEngine()
    # Example of how you might update sysUpTime
    mibBuilder = snmpEngine.get_mib_builder()
    (sysUpTime,) = mibBuilder.import_symbols("__SNMPv2-MIB", "sysUpTime")
    sysUpTime.syntax = TimeTicks(12345)  # Set uptime to 12345

    errorIndication, errorStatus, errorIndex, varBinds = await send_notification(
        snmpEngine,
        CommunityData("public", mpModel=0),
        await UdpTransportTarget.create(("demo.pysnmp.com", 162)),
        ContextData(),
        "trap",
        NotificationType(
            ObjectIdentity("NET-SNMP-EXAMPLES-MIB", "netSnmpExampleNotification")
        ).add_varbinds(
            ObjectType(
                ObjectIdentity("NET-SNMP-EXAMPLES-MIB", "netSnmpExampleHeartbeatRate"),
                1,
            )
        ),
    )

    if errorIndication:
        print(errorIndication)

    snmpEngine.close_dispatcher()


asyncio.run(run())



"""
Bulk walk MIB
+++++++++++++

Send a series of SNMP GETBULK requests using the following options:

* with SNMPv3, user 'usr-none-none', no authentication, no privacy
* over IPv4/UDP
* to an Agent at demo.pysnmp.com:161
* for all OIDs past SNMPv2-MIB::system
* run till end-of-mib condition is reported by Agent
* based on asyncio I/O framework

Functionally similar to:

| $ snmpbulkwalk -v3 -lnoAuthNoPriv -u usr-none-none -Cn0 -Cr50 \
|                demo.pysnmp.com  SNMPv2-MIB::system

"""  #
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run(varBinds):
    snmpEngine = SnmpEngine()
    while True:
        errorIndication, errorStatus, errorIndex, varBindTable = await bulk_cmd(
            snmpEngine,
            UsmUserData("usr-none-none"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
            ContextData(),
            0,
            50,
            *varBinds,
        )

        if errorIndication:
            print(errorIndication)
            break
        elif errorStatus:
            print(
                f"{errorStatus.prettyPrint()} at {varBinds[int(errorIndex) - 1][0] if errorIndex else '?'}"
            )
        else:
            for varBind in varBindTable:
                print(" = ".join([x.prettyPrint() for x in varBind]))

        varBinds = varBindTable
        if is_end_of_mib(varBinds):
            break
    return


asyncio.run(
    run([ObjectType(ObjectIdentity("TCP-MIB")), ObjectType(ObjectIdentity("IP-MIB"))])
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMP-TARGET-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/SNMP-TARGET-MIB
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:25 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#
# IMPORTANT: customization

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(SnmpAdminString,
 SnmpMessageProcessingModel,
 SnmpSecurityLevel,
 SnmpSecurityModel) = mibBuilder.import_symbols(
    "SNMP-FRAMEWORK-MIB",
    "SnmpAdminString",
    "SnmpMessageProcessingModel",
    "SnmpSecurityLevel",
    "SnmpSecurityModel")

(ModuleCompliance,
 NotificationGroup,
 ObjectGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup",
    "ObjectGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 snmpModules) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "snmpModules")

(DisplayString,
 RowStatus,
 StorageType,
 TAddress,
 TDomain,
 TextualConvention,
 TestAndIncr,
 TimeInterval) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "DisplayString",
    "RowStatus",
    "StorageType",
    "TAddress",
    "TDomain",
    "TextualConvention",
    "TestAndIncr",
    "TimeInterval")


# MODULE-IDENTITY

snmpTargetMIB = ModuleIdentity(
    (1, 3, 6, 1, 6, 3, 12)
)
if mibBuilder.loadTexts:
    snmpTargetMIB.setRevisions(
        ("2012-11-27 00:00",
         "2002-10-14 00:00",
         "1998-08-04 00:00",
         "1997-07-14 00:00")
    )
if mibBuilder.loadTexts:
    snmpTargetMIB.setLastUpdated("201211270000Z")
if mibBuilder.loadTexts:
    snmpTargetMIB.setOrganization("IETF SNMPv3 Working Group")
if mibBuilder.loadTexts:
    snmpTargetMIB.setContactInfo("WG-email: snmpv3@lists.tislabs.com Subscribe: majordomo@lists.tislabs.com In message body: subscribe snmpv3 Co-Chair: Russ Mundy Network Associates Laboratories Postal: 15204 Omega Drive, Suite 300 Rockville, MD 20850-4601 USA EMail: mundy@tislabs.com Phone: +1 301-947-7107 Co-Chair: David Harrington Enterasys Networks Postal: 35 Industrial Way P. O. Box 5004 Rochester, New Hampshire 03866-5005 USA EMail: dbh@enterasys.com Phone: +1 603-337-2614 Co-editor: David B. Levi Nortel Networks Postal: 3505 Kesterwood Drive Knoxville, Tennessee 37918 EMail: dlevi@nortelnetworks.com Phone: +1 865 686 0432 Co-editor: Paul Meyer Secure Computing Corporation Postal: 2675 Long Lake Road Roseville, Minnesota 55113 EMail: paul_meyer@securecomputing.com Phone: +1 651 628 1592 Co-editor: Bob Stewart Retired")
if mibBuilder.loadTexts:
    snmpTargetMIB.setDescription("This MIB module defines MIB objects which provide mechanisms to remotely configure the parameters used by an SNMP entity for the generation of SNMP messages. Copyright (C) The Internet Society (2002). This version of this MIB module is part of RFC 3413; see the RFC itself for full legal notices. ")


# Types definitions


# TEXTUAL-CONVENTIONS



class SnmpTagValue(TextualConvention, OctetString):
    status = "current"
    displayHint = "255a"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(0, 255),
    )

    if mibBuilder.loadTexts:
        description = "An octet string containing a tag value. Tag values are preferably in human-readable form. To facilitate internationalization, this information is represented using the ISO/IEC IS 10646-1 character set, encoded as an octet string using the UTF-8 character encoding scheme described in RFC 2279. Since additional code points are added by amendments to the 10646 standard from time to time, implementations must be prepared to encounter any code point from 0x00000000 to 0x7fffffff. The use of control codes should be avoided, and certain control codes are not allowed as described below. For code points not directly supported by user interface hardware or software, an alternative means of entry and display, such as hexadecimal, may be provided. For information encoded in 7-bit US-ASCII, the UTF-8 representation is identical to the US-ASCII encoding. Note that when this TC is used for an object that is used or envisioned to be used as an index, then a SIZE restriction must be specified so that the number of sub-identifiers for any object instance does not exceed the limit of 128, as defined by [RFC1905]. An object of this type contains a single tag value which is used to select a set of entries in a table. A tag value is an arbitrary string of octets, but may not contain a delimiter character. Delimiter characters are defined to be one of the following: - An ASCII space character (0x20). - An ASCII TAB character (0x09). - An ASCII carriage return (CR) character (0x0D). - An ASCII line feed (LF) character (0x0A). Delimiter characters are used to separate tag values in a tag list. An object of this type may only contain a single tag value, and so delimiter characters are not allowed in a value of this type. Some examples of valid tag values are: - 'acme' - 'router' - 'host' The use of a tag value to select table entries is application and MIB specific."

    # IMPORTANT: customization
    encoding = "utf-8"
    _delimiters = (" ", "\n", "\t", "\t")

    def prettyIn(self, value):
        for v in str(value):
            if v in self._delimiters:
                raise error.SmiError("Delimiters not allowed in tag value")
        return OctetString.prettyIn(self, value)

class SnmpTagList(TextualConvention, OctetString):
    status = "current"
    displayHint = "255a"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(0, 255),
    )

    if mibBuilder.loadTexts:
        description = "An octet string containing a list of tag values. Tag values are preferably in human-readable form. To facilitate internationalization, this information is represented using the ISO/IEC IS 10646-1 character set, encoded as an octet string using the UTF-8 character encoding scheme described in RFC 2279. Since additional code points are added by amendments to the 10646 standard from time to time, implementations must be prepared to encounter any code point from 0x00000000 to 0x7fffffff. The use of control codes should be avoided, except as described below. For code points not directly supported by user interface hardware or software, an alternative means of entry and display, such as hexadecimal, may be provided. For information encoded in 7-bit US-ASCII, the UTF-8 representation is identical to the US-ASCII encoding. An object of this type contains a list of tag values which are used to select a set of entries in a table. A tag value is an arbitrary string of octets, but may not contain a delimiter character. Delimiter characters are defined to be one of the following: - An ASCII space character (0x20). - An ASCII TAB character (0x09). - An ASCII carriage return (CR) character (0x0D). - An ASCII line feed (LF) character (0x0A). Delimiter characters are used to separate tag values in a tag list. Only a single delimiter character may occur between two tag values. A tag value may not have a zero length. These constraints imply certain restrictions on the contents of this object: - There cannot be a leading or trailing delimiter character. - There cannot be multiple adjacent delimiter characters. Some examples of valid tag lists are: - An empty string - 'acme router' - 'host managerStation' Note that although a tag value may not have a length of zero, an empty string is still valid. This indicates an empty list (i.e. there are no tag values in the list). The use of the tag list to select table entries is application and MIB specific. Typically, an application will provide one or more tag values, and any entry which contains some combination of these tag values will be selected."

    # IMPORTANT: customization
    encoding = "utf-8"
    _delimiters = (" ", "\n", "\t", "\t")

    def prettyIn(self, value):
        inDelim = True
        for v in str(value):
            if v in self._delimiters:
                if inDelim:
                    raise error.SmiError(
                        "Leading or multiple delimiters not allowed in tag list %r"
                        % value
                    )
                inDelim = True
            else:
                inDelim = False
        if value and inDelim:
            raise error.SmiError(
                "Dangling delimiter not allowed in tag list %r" % value
            )
        return OctetString.prettyIn(self, value)


# MIB Managed Objects in the order of their OIDs

_SnmpTargetObjects_ObjectIdentity = ObjectIdentity
snmpTargetObjects = _SnmpTargetObjects_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 12, 1)
)
_SnmpTargetSpinLock_Type = TestAndIncr
_SnmpTargetSpinLock_Object = MibScalar
snmpTargetSpinLock = _SnmpTargetSpinLock_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 1),
    _SnmpTargetSpinLock_Type()
)
snmpTargetSpinLock.setMaxAccess("read-write")
if mibBuilder.loadTexts:
    snmpTargetSpinLock.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetSpinLock.setDescription("This object is used to facilitate modification of table entries in the SNMP-TARGET-MIB module by multiple managers. In particular, it is useful when modifying the value of the snmpTargetAddrTagList object. The procedure for modifying the snmpTargetAddrTagList object is as follows: 1. Retrieve the value of snmpTargetSpinLock and of snmpTargetAddrTagList. 2. Generate a new value for snmpTargetAddrTagList. 3. Set the value of snmpTargetSpinLock to the retrieved value, and the value of snmpTargetAddrTagList to the new value. If the set fails for the snmpTargetSpinLock object, go back to step 1.")
_SnmpTargetAddrTable_Object = MibTable
snmpTargetAddrTable = _SnmpTargetAddrTable_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2)
)
if mibBuilder.loadTexts:
    snmpTargetAddrTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrTable.setDescription("A table of transport addresses to be used in the generation of SNMP messages.")
_SnmpTargetAddrEntry_Object = MibTableRow
snmpTargetAddrEntry = _SnmpTargetAddrEntry_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1)
)
snmpTargetAddrEntry.setIndexNames(
    (1, "SNMP-TARGET-MIB", "snmpTargetAddrName"),
)
if mibBuilder.loadTexts:
    snmpTargetAddrEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrEntry.setDescription("A transport address to be used in the generation of SNMP operations. Entries in the snmpTargetAddrTable are created and deleted using the snmpTargetAddrRowStatus object.")


class _SnmpTargetAddrName_Type(SnmpAdminString):
    """Custom type snmpTargetAddrName based on SnmpAdminString"""
    subtypeSpec = SnmpAdminString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 32),
    )


_SnmpTargetAddrName_Type.__name__ = "SnmpAdminString"
_SnmpTargetAddrName_Object = MibTableColumn
snmpTargetAddrName = _SnmpTargetAddrName_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 1),
    _SnmpTargetAddrName_Type()
)
snmpTargetAddrName.setMaxAccess("not-accessible")
if mibBuilder.loadTexts:
    snmpTargetAddrName.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrName.setDescription("The locally arbitrary, but unique identifier associated with this snmpTargetAddrEntry.")
_SnmpTargetAddrTDomain_Type = TDomain
_SnmpTargetAddrTDomain_Object = MibTableColumn
snmpTargetAddrTDomain = _SnmpTargetAddrTDomain_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 2),
    _SnmpTargetAddrTDomain_Type()
)
snmpTargetAddrTDomain.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrTDomain.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrTDomain.setDescription("This object indicates the transport type of the address contained in the snmpTargetAddrTAddress object.")
_SnmpTargetAddrTAddress_Type = TAddress
_SnmpTargetAddrTAddress_Object = MibTableColumn
snmpTargetAddrTAddress = _SnmpTargetAddrTAddress_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 3),
    _SnmpTargetAddrTAddress_Type()
)
snmpTargetAddrTAddress.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrTAddress.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrTAddress.setDescription("This object contains a transport address. The format of this address depends on the value of the snmpTargetAddrTDomain object.")


class _SnmpTargetAddrTimeout_Type(TimeInterval):
    """Custom type snmpTargetAddrTimeout based on TimeInterval"""
    defaultValue = 1500


_SnmpTargetAddrTimeout_Type.__name__ = "TimeInterval"
_SnmpTargetAddrTimeout_Object = MibTableColumn
snmpTargetAddrTimeout = _SnmpTargetAddrTimeout_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 4),
    _SnmpTargetAddrTimeout_Type()
)
snmpTargetAddrTimeout.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrTimeout.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrTimeout.setDescription("This object should reflect the expected maximum round trip time for communicating with the transport address defined by this row. When a message is sent to this address, and a response (if one is expected) is not received within this time period, an implementation may assume that the response will not be delivered. Note that the time interval that an application waits for a response may actually be derived from the value of this object. The method for deriving the actual time interval is implementation dependent. One such method is to derive the expected round trip time based on a particular retransmission algorithm and on the number of timeouts which have occurred. The type of message may also be considered when deriving expected round trip times for retransmissions. For example, if a message is being sent with a securityLevel that indicates both authentication and privacy, the derived value may be increased to compensate for extra processing time spent during authentication and encryption processing.")


class _SnmpTargetAddrRetryCount_Type(Integer32):
    """Custom type snmpTargetAddrRetryCount based on Integer32"""
    defaultValue = 3

    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 255),
    )


_SnmpTargetAddrRetryCount_Type.__name__ = "Integer32"
_SnmpTargetAddrRetryCount_Object = MibTableColumn
snmpTargetAddrRetryCount = _SnmpTargetAddrRetryCount_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 5),
    _SnmpTargetAddrRetryCount_Type()
)
snmpTargetAddrRetryCount.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrRetryCount.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrRetryCount.setDescription("This object specifies a default number of retries to be attempted when a response is not received for a generated message. An application may provide its own retry count, in which case the value of this object is ignored.")
_SnmpTargetAddrTagList_Type = SnmpTagList
_SnmpTargetAddrTagList_Object = MibTableColumn
snmpTargetAddrTagList = _SnmpTargetAddrTagList_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 6),
    _SnmpTargetAddrTagList_Type()
)
snmpTargetAddrTagList.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrTagList.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrTagList.setDescription("This object contains a list of tag values which are used to select target addresses for a particular operation.")


class _SnmpTargetAddrParams_Type(SnmpAdminString):
    """Custom type snmpTargetAddrParams based on SnmpAdminString"""
    subtypeSpec = SnmpAdminString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 32),
    )


_SnmpTargetAddrParams_Type.__name__ = "SnmpAdminString"
_SnmpTargetAddrParams_Object = MibTableColumn
snmpTargetAddrParams = _SnmpTargetAddrParams_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 7),
    _SnmpTargetAddrParams_Type()
)
snmpTargetAddrParams.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrParams.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrParams.setDescription("The value of this object identifies an entry in the snmpTargetParamsTable. The identified entry contains SNMP parameters to be used when generating messages to be sent to this transport address.")


class _SnmpTargetAddrStorageType_Type(StorageType):
    """Custom type snmpTargetAddrStorageType based on StorageType"""
    defaultValue = 3


_SnmpTargetAddrStorageType_Type.__name__ = "StorageType"
_SnmpTargetAddrStorageType_Object = MibTableColumn
snmpTargetAddrStorageType = _SnmpTargetAddrStorageType_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 8),
    _SnmpTargetAddrStorageType_Type()
)
snmpTargetAddrStorageType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrStorageType.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrStorageType.setDescription("The storage type for this conceptual row. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
_SnmpTargetAddrRowStatus_Type = RowStatus
_SnmpTargetAddrRowStatus_Object = MibTableColumn
snmpTargetAddrRowStatus = _SnmpTargetAddrRowStatus_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 2, 1, 9),
    _SnmpTargetAddrRowStatus_Type()
)
snmpTargetAddrRowStatus.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetAddrRowStatus.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetAddrRowStatus.setDescription("The status of this conceptual row. To create a row in this table, a manager must set this object to either createAndGo(4) or createAndWait(5). Until instances of all corresponding columns are appropriately configured, the value of the corresponding instance of the snmpTargetAddrRowStatus column is 'notReady'. In particular, a newly created row cannot be made active until the corresponding instances of snmpTargetAddrTDomain, snmpTargetAddrTAddress, and snmpTargetAddrParams have all been set. The following objects may not be modified while the value of this object is active(1): - snmpTargetAddrTDomain - snmpTargetAddrTAddress An attempt to set these objects while the value of snmpTargetAddrRowStatus is active(1) will result in an inconsistentValue error.")
_SnmpTargetParamsTable_Object = MibTable
snmpTargetParamsTable = _SnmpTargetParamsTable_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3)
)
if mibBuilder.loadTexts:
    snmpTargetParamsTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsTable.setDescription("A table of SNMP target information to be used in the generation of SNMP messages.")
_SnmpTargetParamsEntry_Object = MibTableRow
snmpTargetParamsEntry = _SnmpTargetParamsEntry_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1)
)
snmpTargetParamsEntry.setIndexNames(
    (1, "SNMP-TARGET-MIB", "snmpTargetParamsName"),
)
if mibBuilder.loadTexts:
    snmpTargetParamsEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsEntry.setDescription("A set of SNMP target information. Entries in the snmpTargetParamsTable are created and deleted using the snmpTargetParamsRowStatus object.")


class _SnmpTargetParamsName_Type(SnmpAdminString):
    """Custom type snmpTargetParamsName based on SnmpAdminString"""
    subtypeSpec = SnmpAdminString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 32),
    )


_SnmpTargetParamsName_Type.__name__ = "SnmpAdminString"
_SnmpTargetParamsName_Object = MibTableColumn
snmpTargetParamsName = _SnmpTargetParamsName_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 1),
    _SnmpTargetParamsName_Type()
)
snmpTargetParamsName.setMaxAccess("not-accessible")
if mibBuilder.loadTexts:
    snmpTargetParamsName.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsName.setDescription("The locally arbitrary, but unique identifier associated with this snmpTargetParamsEntry.")
_SnmpTargetParamsMPModel_Type = SnmpMessageProcessingModel
_SnmpTargetParamsMPModel_Object = MibTableColumn
snmpTargetParamsMPModel = _SnmpTargetParamsMPModel_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 2),
    _SnmpTargetParamsMPModel_Type()
)
snmpTargetParamsMPModel.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsMPModel.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsMPModel.setDescription("The Message Processing Model to be used when generating SNMP messages using this entry.")


class _SnmpTargetParamsSecurityModel_Type(SnmpSecurityModel):
    """Custom type snmpTargetParamsSecurityModel based on SnmpSecurityModel"""
    subtypeSpec = SnmpSecurityModel.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(1, 2147483647),
    )


_SnmpTargetParamsSecurityModel_Type.__name__ = "SnmpSecurityModel"
_SnmpTargetParamsSecurityModel_Object = MibTableColumn
snmpTargetParamsSecurityModel = _SnmpTargetParamsSecurityModel_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 3),
    _SnmpTargetParamsSecurityModel_Type()
)
snmpTargetParamsSecurityModel.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityModel.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityModel.setDescription("The Security Model to be used when generating SNMP messages using this entry. An implementation may choose to return an inconsistentValue error if an attempt is made to set this variable to a value for a security model which the implementation does not support.")
_SnmpTargetParamsSecurityName_Type = SnmpAdminString
_SnmpTargetParamsSecurityName_Object = MibTableColumn
snmpTargetParamsSecurityName = _SnmpTargetParamsSecurityName_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 4),
    _SnmpTargetParamsSecurityName_Type()
)
snmpTargetParamsSecurityName.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityName.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityName.setDescription("The securityName which identifies the Principal on whose behalf SNMP messages will be generated using this entry.")
_SnmpTargetParamsSecurityLevel_Type = SnmpSecurityLevel
_SnmpTargetParamsSecurityLevel_Object = MibTableColumn
snmpTargetParamsSecurityLevel = _SnmpTargetParamsSecurityLevel_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 5),
    _SnmpTargetParamsSecurityLevel_Type()
)
snmpTargetParamsSecurityLevel.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityLevel.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsSecurityLevel.setDescription("The Level of Security to be used when generating SNMP messages using this entry.")


class _SnmpTargetParamsStorageType_Type(StorageType):
    """Custom type snmpTargetParamsStorageType based on StorageType"""
    defaultValue = 3


_SnmpTargetParamsStorageType_Type.__name__ = "StorageType"
_SnmpTargetParamsStorageType_Object = MibTableColumn
snmpTargetParamsStorageType = _SnmpTargetParamsStorageType_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 6),
    _SnmpTargetParamsStorageType_Type()
)
snmpTargetParamsStorageType.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsStorageType.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsStorageType.setDescription("The storage type for this conceptual row. Conceptual rows having the value 'permanent' need not allow write-access to any columnar objects in the row.")
_SnmpTargetParamsRowStatus_Type = RowStatus
_SnmpTargetParamsRowStatus_Object = MibTableColumn
snmpTargetParamsRowStatus = _SnmpTargetParamsRowStatus_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 3, 1, 7),
    _SnmpTargetParamsRowStatus_Type()
)
snmpTargetParamsRowStatus.setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpTargetParamsRowStatus.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetParamsRowStatus.setDescription("The status of this conceptual row. To create a row in this table, a manager must set this object to either createAndGo(4) or createAndWait(5). Until instances of all corresponding columns are appropriately configured, the value of the corresponding instance of the snmpTargetParamsRowStatus column is 'notReady'. In particular, a newly created row cannot be made active until the corresponding snmpTargetParamsMPModel, snmpTargetParamsSecurityModel, snmpTargetParamsSecurityName, and snmpTargetParamsSecurityLevel have all been set. The following objects may not be modified while the value of this object is active(1): - snmpTargetParamsMPModel - snmpTargetParamsSecurityModel - snmpTargetParamsSecurityName - snmpTargetParamsSecurityLevel An attempt to set these objects while the value of snmpTargetParamsRowStatus is active(1) will result in an inconsistentValue error.")
_SnmpUnavailableContexts_Type = Counter32
_SnmpUnavailableContexts_Object = MibScalar
snmpUnavailableContexts = _SnmpUnavailableContexts_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 4),
    _SnmpUnavailableContexts_Type()
)
snmpUnavailableContexts.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    snmpUnavailableContexts.setStatus("current")
if mibBuilder.loadTexts:
    snmpUnavailableContexts.setDescription("The total number of packets received by the SNMP engine which were dropped because the context contained in the message was unavailable.")
_SnmpUnknownContexts_Type = Counter32
_SnmpUnknownContexts_Object = MibScalar
snmpUnknownContexts = _SnmpUnknownContexts_Object(
    (1, 3, 6, 1, 6, 3, 12, 1, 5),
    _SnmpUnknownContexts_Type()
)
snmpUnknownContexts.setMaxAccess("read-only")
if mibBuilder.loadTexts:
    snmpUnknownContexts.setStatus("current")
if mibBuilder.loadTexts:
    snmpUnknownContexts.setDescription("The total number of packets received by the SNMP engine which were dropped because the context contained in the message was unknown.")
_SnmpTargetConformance_ObjectIdentity = ObjectIdentity
snmpTargetConformance = _SnmpTargetConformance_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 12, 3)
)
_SnmpTargetCompliances_ObjectIdentity = ObjectIdentity
snmpTargetCompliances = _SnmpTargetCompliances_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 12, 3, 1)
)
_SnmpTargetGroups_ObjectIdentity = ObjectIdentity
snmpTargetGroups = _SnmpTargetGroups_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 12, 3, 2)
)

# Managed Objects groups

snmpTargetBasicGroup = ObjectGroup(
    (1, 3, 6, 1, 6, 3, 12, 3, 2, 1)
)
snmpTargetBasicGroup.setObjects(
      *(("SNMP-TARGET-MIB", "snmpTargetSpinLock"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrTDomain"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrTAddress"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrTagList"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrParams"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrStorageType"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrRowStatus"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsMPModel"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsSecurityModel"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsSecurityName"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsSecurityLevel"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsStorageType"),
        ("SNMP-TARGET-MIB", "snmpTargetParamsRowStatus"))
)
if mibBuilder.loadTexts:
    snmpTargetBasicGroup.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetBasicGroup.setDescription("A collection of objects providing basic remote configuration of management targets.")

snmpTargetResponseGroup = ObjectGroup(
    (1, 3, 6, 1, 6, 3, 12, 3, 2, 2)
)
snmpTargetResponseGroup.setObjects(
      *(("SNMP-TARGET-MIB", "snmpTargetAddrTimeout"),
        ("SNMP-TARGET-MIB", "snmpTargetAddrRetryCount"))
)
if mibBuilder.loadTexts:
    snmpTargetResponseGroup.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetResponseGroup.setDescription("A collection of objects providing remote configuration of management targets for applications which generate SNMP messages for which a response message would be expected.")

snmpTargetCommandResponderGroup = ObjectGroup(
    (1, 3, 6, 1, 6, 3, 12, 3, 2, 3)
)
snmpTargetCommandResponderGroup.setObjects(
      *(("SNMP-TARGET-MIB", "snmpUnavailableContexts"),
        ("SNMP-TARGET-MIB", "snmpUnknownContexts"))
)
if mibBuilder.loadTexts:
    snmpTargetCommandResponderGroup.setStatus("current")
if mibBuilder.loadTexts:
    snmpTargetCommandResponderGroup.setDescription("A collection of objects required for command responder applications, used for counting error conditions.")


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance

snmpTargetCommandResponderCompliance = ModuleCompliance(
    (1, 3, 6, 1, 6, 3, 12, 3, 1, 1)
)
snmpTargetCommandResponderCompliance.setObjects(
    ("SNMP-TARGET-MIB", "snmpTargetCommandResponderGroup")
)
if mibBuilder.loadTexts:
    snmpTargetCommandResponderCompliance.setStatus(
        "current"
    )
if mibBuilder.loadTexts:
    snmpTargetCommandResponderCompliance.setDescription("The compliance statement for SNMP entities which include a command responder application.")


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMP-TARGET-MIB",
    **{"SnmpTagValue": SnmpTagValue,
       "SnmpTagList": SnmpTagList,
       "snmpTargetMIB": snmpTargetMIB,
       "snmpTargetObjects": snmpTargetObjects,
       "snmpTargetSpinLock": snmpTargetSpinLock,
       "snmpTargetAddrTable": snmpTargetAddrTable,
       "snmpTargetAddrEntry": snmpTargetAddrEntry,
       "snmpTargetAddrName": snmpTargetAddrName,
       "snmpTargetAddrTDomain": snmpTargetAddrTDomain,
       "snmpTargetAddrTAddress": snmpTargetAddrTAddress,
       "snmpTargetAddrTimeout": snmpTargetAddrTimeout,
       "snmpTargetAddrRetryCount": snmpTargetAddrRetryCount,
       "snmpTargetAddrTagList": snmpTargetAddrTagList,
       "snmpTargetAddrParams": snmpTargetAddrParams,
       "snmpTargetAddrStorageType": snmpTargetAddrStorageType,
       "snmpTargetAddrRowStatus": snmpTargetAddrRowStatus,
       "snmpTargetParamsTable": snmpTargetParamsTable,
       "snmpTargetParamsEntry": snmpTargetParamsEntry,
       "snmpTargetParamsName": snmpTargetParamsName,
       "snmpTargetParamsMPModel": snmpTargetParamsMPModel,
       "snmpTargetParamsSecurityModel": snmpTargetParamsSecurityModel,
       "snmpTargetParamsSecurityName": snmpTargetParamsSecurityName,
       "snmpTargetParamsSecurityLevel": snmpTargetParamsSecurityLevel,
       "snmpTargetParamsStorageType": snmpTargetParamsStorageType,
       "snmpTargetParamsRowStatus": snmpTargetParamsRowStatus,
       "snmpUnavailableContexts": snmpUnavailableContexts,
       "snmpUnknownContexts": snmpUnknownContexts,
       "snmpTargetConformance": snmpTargetConformance,
       "snmpTargetCompliances": snmpTargetCompliances,
       "snmpTargetCommandResponderCompliance": snmpTargetCommandResponderCompliance,
       "snmpTargetGroups": snmpTargetGroups,
       "snmpTargetBasicGroup": snmpTargetBasicGroup,
       "snmpTargetResponseGroup": snmpTargetResponseGroup,
       "snmpTargetCommandResponderGroup": snmpTargetCommandResponderGroup,
       "PYSNMP_MODULE_ID": snmpTargetMIB}
)



from unittest import mock

import pytest

from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.hlapi.v3arch.asyncio.lcd import CommandGeneratorLcdConfigurator


@mock.patch("pysnmp.entity.config.add_v3_user")
@mock.patch("pysnmp.entity.config.delete_v3_user")
@pytest.mark.asyncio
async def test_usm_auth_cache_cleared(delete_v3_user, add_v3_user):
    """
    Ensure auth cache is cleared when auth data is changed.
    """
    with SnmpEngine() as snmpEngine:
        transportTarget = await UdpTransportTarget.create(("198.51.100.1", 161))

        authDataValues = {
            "userName": "username",
            "authKey": "authkey1",
            "authProtocol": USM_AUTH_HMAC96_MD5,
            "privKey": "privkey1",
            "privProtocol": USM_PRIV_CFB128_AES,
        }

        lcd = CommandGeneratorLcdConfigurator()
        initialAuthData = UsmUserData(**authDataValues)
        lcd.configure(snmpEngine, initialAuthData, transportTarget)
        add_v3_user.assert_called_with(
            snmpEngine,
            initialAuthData.userName,
            initialAuthData.authentication_protocol,
            initialAuthData.authentication_key,
            initialAuthData.privacy_protocol,
            initialAuthData.privacy_key,
            securityEngineId=initialAuthData.securityEngineId,
            securityName=initialAuthData.securityName,
            authKeyType=initialAuthData.authKeyType,
            privKeyType=initialAuthData.privKeyType,
        )

        # Ensure we do not add/delete if nothing changes
        add_v3_user.reset_mock()
        lcd.configure(snmpEngine, initialAuthData, transportTarget)
        add_v3_user.assert_not_called()
        delete_v3_user.assert_not_called()

        changeAuthValues = {
            "authKey": "authKey2",
            "privProtocol": USM_PRIV_CBC56_DES,
            "authProtocol": USM_AUTH_HMAC96_SHA,
            "privKey": "privKey2",
        }

        for field, value in changeAuthValues.items():
            add_v3_user.reset_mock()
            delete_v3_user.reset_mock()

            authDataValues[field] = value
            authData = UsmUserData(**authDataValues)
            lcd.configure(snmpEngine, authData, transportTarget)

            delete_v3_user.assert_called_with(
                snmpEngine,
                authData.userName,
                authData.securityEngineId,
            )

            add_v3_user.assert_called_with(
                snmpEngine,
                authData.userName,
                authData.authentication_protocol,
                authData.authentication_key,
                authData.privacy_protocol,
                authData.privacy_key,
                securityEngineId=authData.securityEngineId,
                securityName=authData.securityName,
                authKeyType=authData.authKeyType,
                privKeyType=authData.privKeyType,
            )



#!/usr/bin/env python3
"""
Script to build documentation for the project.
Uses sphinx-polyversion to generate multi-version docs.
"""

import argparse
import os
import subprocess
import sys
from pathlib import Path


def main():
    """Run documentation build using sphinx-polyversion."""
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="Build documentation for this project."
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Run with verbose output"
    )
    args = parser.parse_args()

    print("Building documentation with sphinx-polyversion...", file=sys.stderr)

    # Check if virtual environment is active
    if not Path(".venv").exists():
        print("No virtual environment found. Creating one...", file=sys.stderr)
        subprocess.run(["uv", "venv"], check=True)

        print("Installing dependencies...", file=sys.stderr)
        subprocess.run(["uv", "pip", "install", "-e", ".[dev]"], check=True)

    # Make sure uv and sphinx-polyversion are installed in the venv
    try:
        subprocess.run(["uv", "pip", "install", "uv"], check=True)
        subprocess.run(["uv", "pip", "install", "sphinx-polyversion"], check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error installing dependencies: {e}", file=sys.stderr)
        sys.exit(1)

    # Build the command
    command = ["uv", "run", "sphinx-polyversion", "docs/poly.py"]

    if args.verbose:
        command.append("--verbose")

    # Display the command being run
    print(f"Executing: {' '.join(command)}", file=sys.stderr)

    # Run the command
    try:
        subprocess.run(command, check=True)
        print("Documentation built successfully!", file=sys.stderr)
    except subprocess.CalledProcessError as e:
        print(f"Error building documentation: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import sys
from typing import TYPE_CHECKING


from pyasn1.codec.ber import decoder, eoo
from pyasn1.type import univ
from pysnmp import debug
from pysnmp.proto import errind, error, rfc3411
from pysnmp.proto.api import v1, v2c
from pysnmp.proto.mpmod.base import AbstractMessageProcessingModel
from pysnmp.proto.secmod.base import AbstractSecurityModel

if TYPE_CHECKING:
    from pysnmp.entity.engine import SnmpEngine

# Since I have not found a detailed reference to v1MP/v2cMP
# inner workings, the following has been patterned from v3MP. Most
# references here goes to RFC3412.


class SnmpV1MessageProcessingModel(AbstractMessageProcessingModel):
    """Create a message processing model object."""

    MESSAGE_PROCESSING_MODEL_ID = univ.Integer(0)  # SNMPv1
    SNMP_MSG_SPEC = v1.Message

    # rfc3412: 7.1
    def prepare_outgoing_message(
        self,
        snmpEngine: "SnmpEngine",
        transportDomain,
        transportAddress,
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        pdu,
        expectResponse,
        sendPduHandle,
    ):
        """Prepare SNMP message for dispatch."""
        mibBuilder = snmpEngine.get_mib_builder()

        (snmpEngineId,) = mibBuilder.import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        snmpEngineId = snmpEngineId.syntax

        # rfc3412: 7.1.1b
        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            # serve unique PDU request-id
            msgID = self._cache.new_message_id()
            reqID = pdu.getComponentByPosition(0)
            debug.logger & debug.FLAG_MP and debug.logger(
                f"prepareOutgoingMessage: PDU request-id {reqID} replaced with unique ID {msgID}"
            )

        # rfc3412: 7.1.4
        # Since there's no SNMP engine identification in v1/2c,
        # set destination contextEngineId to ours
        if not contextEngineId:
            contextEngineId = snmpEngineId

        # rfc3412: 7.1.5
        if not contextName:
            contextName = b""

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareOutgoingMessage: using contextEngineId {contextEngineId!r} contextName {contextName!r}"
        )

        # rfc3412: 7.1.6
        scopedPDU = (contextEngineId, contextName, pdu)

        msg = self._snmpMsgSpec
        msg.setComponentByPosition(0, self.MESSAGE_PROCESSING_MODEL_ID)
        msg.setComponentByPosition(2)
        msg.getComponentByPosition(2).setComponentByType(
            pdu.tagSet,
            pdu,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # rfc3412: 7.1.7
        globalData = (msg,)

        k = int(securityModel)
        if k in snmpEngine.security_models:
            smHandler = snmpEngine.security_models[k]
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedSecurityModel
            )

        # rfc3412: 7.1.9.a & rfc2576: 5.2.1 --> no-op

        (snmpEngineMaxMessageSize,) = mibBuilder.import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineMaxMessageSize"
        )

        # fix unique request-id right prior PDU serialization
        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            # noinspection PyUnboundLocalVariable
            pdu.setComponentByPosition(0, msgID)

        # rfc3412: 7.1.9.b
        (securityParameters, wholeMsg) = smHandler.generate_request_message(
            snmpEngine,
            self.MESSAGE_PROCESSING_MODEL_ID,
            globalData,
            snmpEngineMaxMessageSize.syntax,
            securityModel,
            snmpEngineId,
            securityName,
            securityLevel,
            scopedPDU,
        )

        # return original request-id right after PDU serialization
        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            # noinspection PyUnboundLocalVariable
            pdu.setComponentByPosition(0, reqID)

        # rfc3412: 7.1.9.c
        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            # XXX rfc bug? why stateReference should be created?
            self._cache.push_by_message_id(
                int(msgID),
                sendPduHandle=sendPduHandle,
                reqID=reqID,
                snmpEngineId=snmpEngineId,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                transportDomain=transportDomain,
                transportAddress=transportAddress,
            )

        communityName = msg.getComponentByPosition(1)  # for observer

        snmpEngine.observer.store_execution_context(
            snmpEngine,
            "rfc2576.prepareOutgoingMessage",
            dict(
                transportDomain=transportDomain,
                transportAddress=transportAddress,
                wholeMsg=wholeMsg,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                communityName=communityName,
                pdu=pdu,
            ),
        )
        snmpEngine.observer.clear_execution_context(
            snmpEngine, "rfc2576.prepareOutgoingMessage"
        )

        return transportDomain, transportAddress, wholeMsg

    # rfc3412: 7.1
    def prepare_response_message(
        self,
        snmpEngine: "SnmpEngine",
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        pdu,
        maxSizeResponseScopedPDU,
        stateReference,
        statusInformation,
    ):
        """Prepare SNMP message for response."""
        mibBuilder = snmpEngine.get_mib_builder()

        (snmpEngineId,) = mibBuilder.import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        snmpEngineId = snmpEngineId.syntax

        # rfc3412: 7.1.2.b
        if stateReference is None:
            raise error.StatusInformation(errorIndication=errind.nonReportable)

        cachedParams = self._cache.pop_by_state_reference(stateReference)
        msgID = cachedParams["msgID"]
        reqID = cachedParams["reqID"]
        contextEngineId = cachedParams["contextEngineId"]
        contextName = cachedParams["contextName"]
        securityModel = cachedParams["securityModel"]
        securityName = cachedParams["securityName"]
        securityLevel = cachedParams["securityLevel"]
        securityStateReference = cachedParams["securityStateReference"]
        maxMessageSize = cachedParams["msgMaxSize"]
        transportDomain = cachedParams["transportDomain"]
        transportAddress = cachedParams["transportAddress"]

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareResponseMessage: cache read msgID {} transportDomain {} transportAddress {} by stateReference {}".format(
                msgID, transportDomain, transportAddress, stateReference
            )
        )

        # rfc3412: 7.1.3
        if statusInformation:
            # rfc3412: 7.1.3a (N/A)

            # rfc3412: 7.1.3b (always discard)
            raise error.StatusInformation(errorIndication=errind.nonReportable)

        # rfc3412: 7.1.4
        # Since there's no SNMP engine identification in v1/2c,
        # set destination contextEngineId to ours
        if not contextEngineId:
            contextEngineId = snmpEngineId

        # rfc3412: 7.1.5
        if not contextName:
            contextName = b""

        # rfc3412: 7.1.6
        scopedPDU = (contextEngineId, contextName, pdu)

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareResponseMessage: using contextEngineId {contextEngineId!r} contextName {contextName!r}"
        )

        msg = self._snmpMsgSpec
        msg.setComponentByPosition(0, messageProcessingModel)
        msg.setComponentByPosition(2)
        msg.getComponentByPosition(2).setComponentByType(
            pdu.tagSet,
            pdu,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # att: msgId not set back to PDU as it's up to responder app

        # rfc3412: 7.1.7
        globalData = (msg,)

        k = int(securityModel)
        if k in snmpEngine.security_models:
            smHandler = snmpEngine.security_models[k]
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedSecurityModel
            )

        # set original request-id right prior to PDU serialization
        pdu.setComponentByPosition(0, reqID)

        # rfc3412: 7.1.8.a
        (securityParameters, wholeMsg) = smHandler.generate_response_message(
            snmpEngine,
            self.MESSAGE_PROCESSING_MODEL_ID,
            globalData,
            maxMessageSize,
            securityModel,
            snmpEngineId,
            securityName,
            securityLevel,
            scopedPDU,
            securityStateReference,
            statusInformation.get("errorIndication"),
        )

        # recover unique request-id right after PDU serialization
        pdu.setComponentByPosition(0, msgID)

        snmpEngine.observer.store_execution_context(
            snmpEngine,
            "rfc2576.prepareResponseMessage",
            dict(
                transportDomain=transportDomain,
                transportAddress=transportAddress,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                securityEngineId=snmpEngineId,
                communityName=msg.getComponentByPosition(1),
                pdu=pdu,
            ),
        )
        snmpEngine.observer.clear_execution_context(
            snmpEngine, "rfc2576.prepareResponseMessage"
        )

        return transportDomain, transportAddress, wholeMsg

    # rfc3412: 7.2.1

    def prepare_data_elements(
        self, snmpEngine: "SnmpEngine", transportDomain, transportAddress, wholeMsg
    ):
        """Prepare SNMP message data elements."""
        mibBuilder = snmpEngine.get_mib_builder()

        # rfc3412: 7.2.2
        msg, restOfWholeMsg = decoder.decode(wholeMsg, asn1Spec=self._snmpMsgSpec)

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareDataElements: {msg.prettyPrint()}"
        )

        if eoo.endOfOctets.isSameTypeWith(msg):
            raise error.StatusInformation(errorIndication=errind.parseError)

        # rfc3412: 7.2.3
        msgVersion = msg.getComponentByPosition(0)

        # rfc2576: 5.2.1
        (snmpEngineMaxMessageSize,) = mibBuilder.import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineMaxMessageSize"
        )
        communityName = msg.getComponentByPosition(1)
        # transportDomain identifies local endpoint
        securityParameters = (communityName, (transportDomain, transportAddress))
        messageProcessingModel = int(msg.getComponentByPosition(0))
        securityModel = messageProcessingModel + 1
        securityLevel = 1

        # rfc3412: 7.2.4 -- 7.2.5 -> no-op
        smHandler: AbstractSecurityModel
        try:
            try:
                smHandler = snmpEngine.security_models[securityModel]

            except KeyError:
                raise error.StatusInformation(
                    errorIndication=errind.unsupportedSecurityModel
                )

            # rfc3412: 7.2.6
            (
                securityEngineId,
                securityName,
                scopedPDU,
                maxSizeResponseScopedPDU,
                securityStateReference,
            ) = smHandler.process_incoming_message(
                snmpEngine,
                messageProcessingModel,
                snmpEngineMaxMessageSize.syntax,
                securityParameters,
                securityModel,
                securityLevel,
                wholeMsg,
                msg,
            )  # type: ignore

            debug.logger & debug.FLAG_MP and debug.logger(
                f"prepareDataElements: SM returned securityEngineId {securityEngineId!r} securityName {securityName!r}"
            )

        except error.StatusInformation:
            statusInformation = sys.exc_info()[1]

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc2576.prepareDataElements:sm-failure",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityLevel=securityLevel,
                    securityParameters=securityParameters,
                    statusInformation=statusInformation,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc2576.prepareDataElements:sm-failure"
            )

            raise

        # rfc3412: 7.2.6a --> no-op

        # rfc3412: 7.2.7
        contextEngineId, contextName, pdu = scopedPDU

        # rfc2576: 5.2.1
        pduVersion = msgVersion
        pduType = pdu.tagSet

        # rfc3412: 7.2.8, 7.2.9 -> no-op

        # rfc3412: 7.2.10
        if pduType in rfc3411.RESPONSE_CLASS_PDUS:
            # get unique PDU request-id
            msgID = pdu.getComponentByPosition(0)

            # 7.2.10a
            try:
                cachedReqParams = self._cache.pop_by_message_id(int(msgID))
            except error.ProtocolError:
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.dataMismatch)

            # recover original PDU request-id to return to app
            pdu.setComponentByPosition(0, cachedReqParams["reqID"])

            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareDataElements: unique PDU request-id {} replaced with original ID {}".format(
                    msgID, cachedReqParams["reqID"]
                )
            )

            # 7.2.10b
            sendPduHandle = cachedReqParams["sendPduHandle"]
        else:
            sendPduHandle = None

        # no error by default
        statusInformation = None

        # rfc3412: 7.2.11 -> no-op

        # rfc3412: 7.2.12
        if pduType in rfc3411.RESPONSE_CLASS_PDUS:
            # rfc3412: 7.2.12a -> no-op
            # rfc3412: 7.2.12b
            # noinspection PyUnboundLocalVariable
            if (
                securityModel != cachedReqParams["securityModel"]
                or securityName != cachedReqParams["securityName"]
                or securityLevel != cachedReqParams["securityLevel"]
                or contextEngineId != cachedReqParams["contextEngineId"]
                or contextName != cachedReqParams["contextName"]
            ):
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.dataMismatch)

            stateReference = None

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc2576.prepareDataElements:response",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    communityName=communityName,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc2576.prepareDataElements:response"
            )

            # rfc3412: 7.2.12c
            smHandler.release_state_information(securityStateReference)

            # rfc3412: 7.2.12d
            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        # rfc3412: 7.2.13
        if pduType in rfc3411.CONFIRMED_CLASS_PDUS:
            # store original PDU request-id and replace it with a unique one
            reqID = pdu.getComponentByPosition(0)
            msgID = self._cache.new_message_id()
            pdu.setComponentByPosition(0, msgID)
            debug.logger & debug.FLAG_MP and debug.logger(
                f"prepareDataElements: received PDU request-id {reqID} replaced with unique ID {msgID}"
            )

            # rfc3412: 7.2.13a
            (snmpEngineId,) = mibBuilder.import_symbols(
                "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
            )
            if securityEngineId != snmpEngineId.syntax:
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.engineIDMismatch)

            # rfc3412: 7.2.13b
            stateReference = self._cache.new_state_reference()
            self._cache.push_by_state_reference(
                stateReference,
                msgVersion=messageProcessingModel,
                msgID=msgID,
                reqID=reqID,
                contextEngineId=contextEngineId,
                contextName=contextName,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                securityStateReference=securityStateReference,
                msgMaxSize=snmpEngineMaxMessageSize.syntax,
                maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                transportDomain=transportDomain,
                transportAddress=transportAddress,
            )

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc2576.prepareDataElements:confirmed",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    communityName=communityName,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc2576.prepareDataElements:confirmed"
            )

            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareDataElements: cached by new stateReference %s" % stateReference
            )

            # rfc3412: 7.2.13c
            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        # rfc3412: 7.2.14
        if pduType in rfc3411.UNCONFIRMED_CLASS_PDUS:
            # Pass new stateReference to let app browse request details
            stateReference = self._cache.new_state_reference()

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc2576.prepareDataElements:unconfirmed",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    communityName=communityName,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc2576.prepareDataElements:unconfirmed"
            )

            # This is not specified explicitly in RFC
            smHandler.release_state_information(securityStateReference)

            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        smHandler.release_state_information(securityStateReference)
        raise error.StatusInformation(errorIndication=errind.unsupportedPDUtype)


class SnmpV2cMessageProcessingModel(SnmpV1MessageProcessingModel):
    """Create a message processing model object."""

    MESSAGE_PROCESSING_MODEL_ID = univ.Integer(1)  # SNMPv2c
    SNMP_MSG_SPEC = v2c.Message



"""
Broadcast SNMP message (IPv4)
+++++++++++++++++++++++++++++

Send SNMP GET request to broadcast address and wait for respons(es):

* with SNMPv2c, community 'public'
* over IPv4/UDP
* to all Agents via broadcast address 255.255.255.255:161
* for OIDs in tuple form

Here we send out a single SNMP request and wait for potentially many SNMP
responses from multiple SNMP Agents listening in local broadcast domain.
Since we can't predict the exact number of Agents responding, this script
just waits for arbitrary time for collecting all responses. This technology
is also known as SNMP-based discovery.

This script performs similar to the following Net-SNMP command:

| $ snmpget -v2c -c public -ObentU 255.255.255.255 1.3.6.1.2.1.1.1.0 1.3.6.1.2.1.1.3.0

"""  #
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.carrier.asyncio.dgram import udp
from pyasn1.codec.ber import encoder, decoder
from pysnmp.proto import api

# Broadcast manager settings
maxWaitForResponses = 5
maxNumberResponses = 10

# Protocol version to use
# pMod = api.PROTOCOL_MODULES[api.SNMP_VERSION_1]
pMod = api.PROTOCOL_MODULES[api.SNMP_VERSION_2C]

# Build PDU
reqPDU = pMod.GetRequestPDU()
pMod.apiPDU.set_defaults(reqPDU)
pMod.apiPDU.set_varbinds(
    reqPDU, (("1.3.6.1.2.1.1.1.0", pMod.Null("")), ("1.3.6.1.2.1.1.3.0", pMod.Null("")))
)

# Build message
reqMsg = pMod.Message()
pMod.apiMessage.set_defaults(reqMsg)
pMod.apiMessage.set_community(reqMsg, "public")
pMod.apiMessage.set_pdu(reqMsg, reqPDU)


# noinspection PyUnusedLocal,PyUnusedLocal
def __callback(
    transportDispatcher, transportDomain, transportAddress, wholeMsg, reqPDU=reqPDU
):
    while wholeMsg:
        rspMsg, wholeMsg = decoder.decode(wholeMsg, asn1Spec=pMod.Message())
        rspPDU = pMod.apiMessage.get_pdu(rspMsg)
        # Match response to request
        if pMod.apiPDU.get_request_id(reqPDU) == pMod.apiPDU.get_request_id(rspPDU):
            # Check for SNMP errors reported
            errorStatus = pMod.apiPDU.get_error_status(rspPDU)
            if errorStatus:
                print(errorStatus.prettyPrint())
            else:
                for oid, val in pMod.apiPDU.get_varbinds(rspPDU):
                    print(f"{oid.prettyPrint()} = {val.prettyPrint()}")
            transportDispatcher.job_finished(1)
    return wholeMsg


transportDispatcher = AsyncioDispatcher()

transportDispatcher.register_recv_callback(__callback)

# UDP/IPv4
udpSocketTransport = udp.UdpAsyncioTransport().open_client_mode(allow_broadcast=True)
transportDispatcher.register_transport(udp.DOMAIN_NAME, udpSocketTransport)

# Pass message to dispatcher
transportDispatcher.send_message(
    encoder.encode(reqMsg), udp.DOMAIN_NAME, ("255.255.255.255", 161)
)

# wait for a maximum of 10 responses or time out
transportDispatcher.job_started(1, maxNumberResponses)

# Dispatcher will finish as all jobs counter reaches zero
try:
    transportDispatcher.run_dispatcher(maxWaitForResponses)
except:
    raise
finally:
    transportDispatcher.close_dispatcher()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import random
from hashlib import md5, sha1

from pyasn1.type import univ
from pysnmp.proto import errind, error
from pysnmp.proto.secmod.rfc3414 import localkey
from pysnmp.proto.secmod.rfc3414.auth import hmacmd5, hmacsha
from pysnmp.proto.secmod.rfc3414.priv import base
from pysnmp.proto.secmod.rfc7860.auth import hmacsha2

PysnmpCryptoError = False
try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.decrepit.ciphers import algorithms
    from cryptography.hazmat.primitives.ciphers import Cipher, modes
except ImportError:
    PysnmpCryptoError = True

random.seed()


# 5.1.1


class Des3(base.AbstractEncryptionService):
    """Reeder 3DES-EDE for USM (Internet draft).

    https://tools.ietf.org/html/draft-reeder-snmpv3-usm-3desede-00
    """

    SERVICE_ID = (1, 3, 6, 1, 6, 3, 10, 1, 2, 3)  # usm3DESEDEPrivProtocol
    KEY_SIZE = 32
    local_int = random.randrange(0, 0xFFFFFFFF)

    def hash_passphrase(self, authProtocol, privKey) -> univ.OctetString:
        """Hash a passphrase.

        This method hashes a passphrase using the authentication protocol.

        Parameters
        ----------
        authProtocol : tuple
            The authentication protocol OID.
        privKey : bytes
            The privacy key.

        Returns
        -------
        bytes
            The hashed passphrase.
        """
        if authProtocol == hmacmd5.HmacMd5.SERVICE_ID:
            hashAlgo = md5
        elif authProtocol == hmacsha.HmacSha.SERVICE_ID:
            hashAlgo = sha1
        elif authProtocol in hmacsha2.HmacSha2.HASH_ALGORITHM:
            hashAlgo = hmacsha2.HmacSha2.HASH_ALGORITHM[authProtocol]
        else:
            raise error.ProtocolError(f"Unknown auth protocol {authProtocol}")
        return localkey.hash_passphrase(privKey, hashAlgo)

    # 2.1
    def localize_key(self, authProtocol, privKey, snmpEngineID) -> bytes:
        """3-DES key localization algorithm.

        This algorithm is used to localize a 3-DES key to an authoritative
        engine ID.

        Parameters
        ----------
        authProtocol : tuple
            The authentication protocol OID.
        privKey : bytes
            The privacy key.
        snmpEngineID : bytes
            The authoritative engine ID.

        Returns
        -------
        bytes
            The localized key.
        """
        if authProtocol == hmacmd5.HmacMd5.SERVICE_ID:
            hashAlgo = md5
        elif authProtocol == hmacsha.HmacSha.SERVICE_ID:
            hashAlgo = sha1
        elif authProtocol in hmacsha2.HmacSha2.HASH_ALGORITHM:
            hashAlgo = hmacsha2.HmacSha2.HASH_ALGORITHM[authProtocol]
        else:
            raise error.ProtocolError(f"Unknown auth protocol {authProtocol}")
        localPrivKey = localkey.localize_key(privKey, snmpEngineID, hashAlgo)

        # now extend this key if too short by repeating steps that includes the hashPassphrase step
        while len(localPrivKey) < self.KEY_SIZE:
            # this is the difference between reeder and bluementhal
            newKey = localkey.hash_passphrase(localPrivKey, hashAlgo)
            localPrivKey += localkey.localize_key(newKey, snmpEngineID, hashAlgo)

        return localPrivKey[: self.KEY_SIZE]

    # 5.1.1.1
    def __get_encryption_key(self, privKey, snmpEngineBoots):
        # 5.1.1.1.1
        des3Key = privKey[:24]
        preIV = privKey[24:32]

        securityEngineBoots = int(snmpEngineBoots)

        salt = [
            securityEngineBoots >> 24 & 0xFF,
            securityEngineBoots >> 16 & 0xFF,
            securityEngineBoots >> 8 & 0xFF,
            securityEngineBoots & 0xFF,
            self.local_int >> 24 & 0xFF,
            self.local_int >> 16 & 0xFF,
            self.local_int >> 8 & 0xFF,
            self.local_int & 0xFF,
        ]
        if self.local_int == 0xFFFFFFFF:
            self.local_int = 0
        else:
            self.local_int += 1

        # salt not yet hashed XXX

        return (
            des3Key.asOctets(),
            univ.OctetString(salt).asOctets(),
            univ.OctetString(
                map(lambda x, y: x ^ y, salt, preIV.asNumbers())
            ).asOctets(),
        )

    @staticmethod
    def __get_decryption_key(privKey, salt):
        return (
            privKey[:24].asOctets(),
            univ.OctetString(
                map(lambda x, y: x ^ y, salt.asNumbers(), privKey[24:32].asNumbers())
            ).asOctets(),
        )

    # 5.1.1.2
    def encrypt_data(self, encryptKey, privParameters, dataToEncrypt):
        """Encrypt data."""
        if PysnmpCryptoError:
            raise error.StatusInformation(errorIndication=errind.encryptionError)

        snmpEngineBoots, snmpEngineTime, salt = privParameters

        des3Key, salt, iv = self.__get_encryption_key(encryptKey, snmpEngineBoots)

        privParameters = univ.OctetString(salt)

        plaintext = (
            dataToEncrypt
            + univ.OctetString((0,) * (8 - len(dataToEncrypt) % 8)).asOctets()
        )

        try:
            des3 = _cryptography_cipher(des3Key, iv).encryptor()
            ciphertext = des3.update(plaintext) + des3.finalize()

        except AttributeError:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedPrivProtocol
            )

        return univ.OctetString(ciphertext), privParameters

    # 5.1.1.3
    def decrypt_data(self, decryptKey, privParameters, encryptedData):
        """Decrypt data."""
        if PysnmpCryptoError:
            raise error.StatusInformation(errorIndication=errind.decryptionError)
        snmpEngineBoots, snmpEngineTime, salt = privParameters

        if len(salt) != 8:
            raise error.StatusInformation(errorIndication=errind.decryptionError)

        des3Key, iv = self.__get_decryption_key(decryptKey, salt)

        if len(encryptedData) % 8 != 0:
            raise error.StatusInformation(errorIndication=errind.decryptionError)

        ciphertext = encryptedData.asOctets()

        try:
            des3 = _cryptography_cipher(des3Key, iv).decryptor()
            plaintext = des3.update(ciphertext) + des3.finalize()

        except AttributeError:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedPrivProtocol
            )

        return plaintext


def _cryptography_cipher(key, iv):
    """Build a cryptography TripleDES Cipher object.

    :param bytes key: Encryption key
    :param bytesiv iv: Initialization vector
    :returns: TripleDES Cipher instance
    :rtype: cryptography.hazmat.primitives.ciphers.Cipher
    """
    return Cipher(
        algorithm=algorithms.TripleDES(key),
        mode=modes.CBC(iv),
        backend=default_backend(),
    )



import asyncio
import pytest


from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.api import v2c
from tests.manager_context import MANAGER_PORT, ManagerContextManager


@pytest.mark.asyncio
async def test_send_trap_enterprise_specific():
    async with ManagerContextManager() as (_, message_count):
        with SnmpEngine() as snmpEngine:
            (
                errorIndication,
                errorStatus,
                errorIndex,
                varBinds,
            ) = await send_notification(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", MANAGER_PORT)),
                ContextData(),
                "trap",
                NotificationType(
                    ObjectIdentity("1.3.6.1.4.1.20408.4.1.1.2.432")
                ).add_varbinds(
                    (v2c.apiTrapPDU.sysUpTime, TimeTicks(12345)),
                    ("1.3.6.1.2.1.1.1.0", OctetString("my system")),
                    (v2c.apiTrapPDU.snmpTrapAddress, IpAddress("127.0.0.1")),
                    (
                        ObjectIdentity("SNMPv2-MIB", "snmpTrapOID", 0),
                        ObjectIdentifier("1.3.6.1.4.1.20408.4.1.1.2.432"),
                    ),
                ),
            )

        await asyncio.sleep(1)
        assert message_count == [1]


@pytest.mark.asyncio
async def test_send_trap_generic():
    async with ManagerContextManager() as (_, message_count):
        with SnmpEngine() as snmpEngine:
            (
                errorIndication,
                errorStatus,
                errorIndex,
                varBinds,
            ) = await send_notification(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", MANAGER_PORT)),
                ContextData(),
                "trap",
                NotificationType(ObjectIdentity("1.3.6.1.6.3.1.1.5.2"))
                .load_mibs("SNMPv2-MIB")
                .add_varbinds(
                    (
                        "1.3.6.1.6.3.1.1.4.3.0",
                        "1.3.6.1.4.1.20408.4.1.1.2",
                    ),  # IMPORTANT: MIB is needed to resolve str to correct type.
                    ("1.3.6.1.2.1.1.1.0", OctetString("my system")),
                ),
            )

        await asyncio.sleep(1)
        assert message_count == [1]


@pytest.mark.asyncio
async def test_send_trap_custom_mib():
    async with ManagerContextManager() as (_, message_count):
        with SnmpEngine() as snmpEngine:
            mibBuilder = snmpEngine.get_mib_builder()
            (sysUpTime,) = mibBuilder.import_symbols("__SNMPv2-MIB", "sysUpTime")
            sysUpTime.syntax = TimeTicks(12345)

            (
                errorIndication,
                errorStatus,
                errorIndex,
                varBinds,
            ) = await send_notification(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", MANAGER_PORT)),
                ContextData(),
                "trap",
                NotificationType(
                    ObjectIdentity(
                        "NET-SNMP-EXAMPLES-MIB", "netSnmpExampleNotification"
                    )
                ).add_varbinds(
                    ObjectType(
                        ObjectIdentity(
                            "NET-SNMP-EXAMPLES-MIB", "netSnmpExampleHeartbeatRate"
                        ),
                        1,
                    )
                ),
            )

        await asyncio.sleep(1)
        assert message_count == [1]



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.errind import DecryptionError, UnknownUserName, WrongDigest
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_usm_sha_aes256():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            authData = UsmUserData(
                "usr-sha-aes256",
                "authkey1",
                "privkey1",
                authProtocol=USM_AUTH_HMAC96_SHA,
                privProtocol=USM_PRIV_CFB256_AES,
            )
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                authData,
                await UdpTransportTarget.create(("localhost", AGENT_PORT), retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_usm_sha_aes256_wrong_auth():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            authData = UsmUserData(
                "usr-sha-aes",
                "authkey1",
                "privkey1",
                authProtocol=USM_AUTH_HMAC96_MD5,  # wrongly use usmHMACMD5AuthProtocol
                privProtocol=USM_PRIV_CFB256_AES,
            )
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                authData,
                await UdpTransportTarget.create(("localhost", AGENT_PORT), retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert isinstance(errorIndication, WrongDigest)
            assert str(errorIndication) == "Wrong SNMP PDU digest"


@pytest.mark.asyncio
async def test_usm_sha_aes256_wrong_user():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            authData = UsmUserData(
                "usr-sha-aes-not-exist",
                "authkey1",
                "privkey1",
                authProtocol=USM_AUTH_HMAC96_SHA,
                privProtocol=USM_PRIV_CFB256_AES,
            )
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                authData,
                await UdpTransportTarget.create(("localhost", AGENT_PORT), retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert isinstance(errorIndication, UnknownUserName)
            assert str(errorIndication) == "Unknown USM user"



#!/usr/bin/env python3
"""
Script to bump version numbers for pysmi.
Uses bump2version to increment version.
"""

import argparse
import subprocess
import sys


def main():
    """Main function to run bump2version with different options."""
    parser = argparse.ArgumentParser(description="Bump version for the project.")

    # Version part to bump (mutually exclusive options)
    part_group = parser.add_mutually_exclusive_group(required=True)
    part_group.add_argument("--major", action="store_true", help="Bump major version")
    part_group.add_argument("--minor", action="store_true", help="Bump minor version")
    part_group.add_argument("--patch", action="store_true", help="Bump patch version")

    # Optional flags
    parser.add_argument(
        "--dry-run", action="store_true", help="Don't actually change any files"
    )
    parser.add_argument(
        "--tag", action="store_true", help="Create a git tag for the new version"
    )
    parser.add_argument(
        "--tag-message", metavar="MSG", help="Tag message (default: version number)"
    )

    args = parser.parse_args()

    # Determine which part to bump
    bump_part = "patch"  # Default
    if args.major:
        bump_part = "major"
    elif args.minor:
        bump_part = "minor"

    # Build the command
    cmd = ["bump2version"]

    # Add appropriate flags
    cmd.append("--allow-dirty")

    if args.dry_run:
        cmd.append("--dry-run")

    if args.tag:
        cmd.append("--tag")

    if args.tag_message:
        cmd.extend(["--tag-message", args.tag_message])

    # Add the part to bump
    cmd.append(bump_part)

    # Execute the bump2version command
    print(f"Executing: {' '.join(cmd)}")
    try:
        result = subprocess.run(cmd, check=True, capture_output=True, text=True)
        print(result.stdout)

        if result.stderr:
            print("Errors/Warnings:", file=sys.stderr)
            print(result.stderr, file=sys.stderr)
    except subprocess.CalledProcessError as e:
        print(f"Error: {e}", file=sys.stderr)
        print(e.stderr, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# Copyright (C) 2014, Zebra Technologies
# Authors: Matt Hooks <me@matthooks.com>
#          Zachary Lorusso <zlorusso@gmail.com>
#
import asyncio
import warnings


from pysnmp.entity.engine import SnmpEngine
from pysnmp.entity.rfc3413 import ntforg
from pysnmp.hlapi.transport import AbstractTransportTarget
from pysnmp.hlapi.v3arch.asyncio.auth import CommunityData, UsmUserData
from pysnmp.hlapi.v3arch.asyncio.context import ContextData
from pysnmp.hlapi.v3arch.asyncio.lcd import NotificationOriginatorLcdConfigurator
from pysnmp.hlapi.varbinds import NotificationOriginatorVarBinds
from pysnmp.smi.rfc1902 import NotificationType

__all__ = ["send_notification"]

VB_PROCESSOR = NotificationOriginatorVarBinds()
LCD = NotificationOriginatorLcdConfigurator()


async def send_notification(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    notifyType: str,
    *varBinds: NotificationType,
    **options,
):
    r"""Creates a generator to send SNMP notification.

    When iterator gets advanced by :py:mod:`asyncio` main loop,
    SNMP TRAP or INFORM notification is send (:RFC:`1905#section-4.2.6`).
    The iterator yields :py:class:`asyncio.get_running_loop().create_future()` which gets done whenever
    response arrives or error occurs.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    notifyType : str
        Indicates type of notification to be sent. Recognized literal
        values are *trap* or *inform*.

    \*varBinds: :class:`tuple` of OID-value pairs or :py:class:`~pysnmp.smi.rfc1902.ObjectType` or :py:class:`~pysnmp.smi.rfc1902.NotificationType`
        One or more objects representing MIB variables to place
        into SNMP notification. It could be tuples of OID-values
        or :py:class:`~pysnmp.smi.rfc1902.ObjectType` class instances
        of :py:class:`~pysnmp.smi.rfc1902.NotificationType` objects.

        SNMP Notification PDU places rigid requirement on the ordering of
        the variable-bindings.

        Mandatory variable-bindings:

        0. SNMPv2-MIB::sysUpTime.0 = <agent uptime>
        1. SNMPv2-SMI::snmpTrapOID.0 = {SNMPv2-MIB::coldStart, ...}

        Optional variable-bindings (applicable to SNMP v1 TRAP):

        2. SNMP-COMMUNITY-MIB::snmpTrapAddress.0 = <agent-IP>
        3. SNMP-COMMUNITY-MIB::snmpTrapCommunity.0 = <snmp-community-name>
        4. SNMP-COMMUNITY-MIB::snmpTrapEnterprise.0 = <enterprise-OID>

        Informational variable-bindings:

        * SNMPv2-SMI::NOTIFICATION-TYPE
        * SNMPv2-SMI::OBJECT-TYPE

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.

    Yields
    ------
    errorIndication : :py:class:`~pysnmp.proto.errind.ErrorIndication`
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBinds : tuple
        A sequence of :py:class:`~pysnmp.smi.rfc1902.ObjectType` class
        instances representing MIB variables returned in SNMP response.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Examples
    --------
    >>> import asyncio
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>>
    >>> async def run():
    ...     errorIndication, errorStatus, errorIndex, varBinds = await send_notification(
    ...         SnmpEngine(),
    ...         CommunityData('public'),
    ...         await UdpTransportTarget.create(('demo.pysnmp.com', 162)),
    ...         ContextData(),
    ...         'trap',
    ...         NotificationType(ObjectIdentity('IF-MIB', 'linkDown')))
    ...     print(errorIndication, errorStatus, errorIndex, varBinds)
    ...
    >>> asyncio.run(run())
    (None, 0, 0, [])
    >>>

    """

    def __callback(
        snmpEngine: SnmpEngine,
        sendRequestHandle,
        errorIndication,
        errorStatus,
        errorIndex,
        varBinds,
        cbCtx,
    ):
        lookupMib, future = cbCtx
        if future.cancelled():
            return
        try:
            varBindsUnmade = VB_PROCESSOR.unmake_varbinds(
                snmpEngine.cache, varBinds, lookupMib
            )
        except Exception as e:
            future.set_exception(e)
        else:
            future.set_result(
                (errorIndication, errorStatus, errorIndex, varBindsUnmade)
            )

    notifyName = LCD.configure(
        snmpEngine, authData, transportTarget, notifyType, contextData.contextName
    )

    future = asyncio.get_running_loop().create_future()

    ntforg.NotificationOriginator().send_varbinds(
        snmpEngine,
        notifyName,
        contextData.contextEngineId,
        contextData.contextName,
        VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds),
        __callback,
        (options.get("lookupMib", True), future),
    )

    if notifyType == "trap":

        def __trap_function(future):
            if future.cancelled():
                return
            future.set_result((None, 0, 0, []))

        loop = asyncio.get_event_loop()
        loop.call_soon(__trap_function, future)

    return await future


# Compatibility API
deprecated_attributes = {
    "sendNotification": "send_notification",
}


def __getattr__(attr: str):
    if new_attr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {new_attr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[new_attr]
    raise AttributeError(f"module '{__name__}' has no attribute '{attr}'")



"""
SNMPv3, auth: SHA, privacy: AES128
++++++++++++++++++++++++++++++++++

Send a SNMP GET request
* with SNMPv3 with user 'usr-sha-aes', SHA auth and AES128 privacy protocols
* over IPv4/UDP
* to an Agent at 127.0.0.1:161
* for an OID in tuple form

This script performs similar to the following Net-SNMP command:

| $ snmpget -v3 -l authPriv -u usr-sha-aes -a SHA -A authkey1 -x AES -X privkey1 -ObentU 127.0.0.1:161  1.3.6.1.2.1.1.1.0

"""  #

from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import cmdgen

# Create SNMP engine instance
snmpEngine = engine.SnmpEngine()

#
# SNMPv3/USM setup
#

# user: usr-sha-aes, auth: SHA, priv AES
config.add_v3_user(
    snmpEngine,
    "usr-sha-aes",
    config.USM_AUTH_HMAC96_SHA,
    "authkey1",
    config.USM_PRIV_CFB128_AES,
    "privkey1",
)
config.add_target_parameters(snmpEngine, "my-creds", "usr-sha-aes", "authPriv")

#
# Setup transport endpoint and bind it with security settings yielding
# a target name
#

# UDP/IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)
config.add_target_address(
    snmpEngine, "my-router", udp.DOMAIN_NAME, ("127.0.0.1", 161), "my-creds"
)


# Error/response receiver
# noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
def cbFun(
    snmpEngine,
    sendRequestHandle,
    errorIndication,
    errorStatus,
    errorIndex,
    varBinds,
    cbCtx,
):
    if errorIndication:
        print(errorIndication)
    elif errorStatus:
        print(
            f"{errorStatus.prettyPrint()} at {varBinds[int(errorIndex) - 1][0] if errorIndex else '?'}"
        )
    else:
        for oid, val in varBinds:
            print(f"{oid.prettyPrint()} = {val.prettyPrint()}")


# Prepare and send a request message
cmdgen.GetCommandGenerator().send_varbinds(
    snmpEngine,
    "my-router",
    None,
    "",  # contextEngineId, contextName
    [((1, 3, 6, 1, 2, 1, 1, 1, 0), None)],
    cbFun,
)

# Run I/O dispatcher which would send pending queries and process responses
snmpEngine.open_dispatcher(3)

snmpEngine.close_dispatcher()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from typing import Any

from pysnmp.entity import config
from pysnmp.entity.engine import SnmpEngine
from pysnmp.error import PySnmpError
from pysnmp.smi.error import NoSuchInstanceError, SmiError


def get_target_address(snmpEngine: SnmpEngine, snmpTargetAddrName):
    """Return transport endpoint information for a given target."""
    mibBuilder = snmpEngine.get_mib_builder()

    (snmpTargetAddrEntry,) = mibBuilder.import_symbols(  # type: ignore
        "SNMP-TARGET-MIB", "snmpTargetAddrEntry"
    )

    cache: "dict[str, Any] | None" = snmpEngine.get_user_context("getTargetAddr")
    if cache is None:
        cache = {"id": -1}
        snmpEngine.set_user_context(getTargetAddr=cache)

    if cache["id"] != snmpTargetAddrEntry.branchVersionId:
        cache["nameToTargetMap"] = {}

    nameToTargetMap = cache["nameToTargetMap"]

    if snmpTargetAddrName not in nameToTargetMap:
        (
            snmpTargetAddrTDomain,
            snmpTargetAddrTAddress,
            snmpTargetAddrTimeout,
            snmpTargetAddrRetryCount,
            snmpTargetAddrParams,
        ) = mibBuilder.import_symbols(  # type: ignore
            "SNMP-TARGET-MIB",
            "snmpTargetAddrTDomain",
            "snmpTargetAddrTAddress",
            "snmpTargetAddrTimeout",
            "snmpTargetAddrRetryCount",
            "snmpTargetAddrParams",
        )
        (snmpSourceAddrTAddress,) = mibBuilder.import_symbols("PYSNMP-SOURCE-MIB", "snmpSourceAddrTAddress")  # type: ignore

        tblIdx = snmpTargetAddrEntry.getInstIdFromIndices(snmpTargetAddrName)

        try:
            snmpTargetAddrTDomain = snmpTargetAddrTDomain.getNode(
                snmpTargetAddrTDomain.name + tblIdx
            ).syntax
            snmpTargetAddrTAddress = snmpTargetAddrTAddress.getNode(
                snmpTargetAddrTAddress.name + tblIdx
            ).syntax
            snmpTargetAddrTimeout = snmpTargetAddrTimeout.getNode(
                snmpTargetAddrTimeout.name + tblIdx
            ).syntax
            snmpTargetAddrRetryCount = snmpTargetAddrRetryCount.getNode(
                snmpTargetAddrRetryCount.name + tblIdx
            ).syntax
            snmpTargetAddrParams = snmpTargetAddrParams.getNode(
                snmpTargetAddrParams.name + tblIdx
            ).syntax
            snmpSourceAddrTAddress = snmpSourceAddrTAddress.getNode(
                snmpSourceAddrTAddress.name + tblIdx
            ).syntax
        except NoSuchInstanceError:
            raise SmiError("Target %s not configured to LCD" % snmpTargetAddrName)

        if snmpEngine.transport_dispatcher is None:
            raise PySnmpError("TransportDispatcher not set")
        transport = snmpEngine.transport_dispatcher.get_transport(snmpTargetAddrTDomain)

        if (
            snmpTargetAddrTDomain[: len(config.SNMP_UDP_DOMAIN)]
            == config.SNMP_UDP_DOMAIN
        ):
            (
                SnmpUDPAddress,
            ) = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
                "SNMPv2-TM", "SnmpUDPAddress"
            )
            addr = transport.ADDRESS_TYPE(  # type: ignore
                SnmpUDPAddress(snmpTargetAddrTAddress)
            ).set_local_address(SnmpUDPAddress(snmpSourceAddrTAddress))
        elif (
            snmpTargetAddrTDomain[: len(config.SNMP_UDP6_DOMAIN)]
            == config.SNMP_UDP6_DOMAIN
        ):
            (TransportAddressIPv6,) = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
                "TRANSPORT-ADDRESS-MIB", "TransportAddressIPv6"
            )
            addr = transport.ADDRESS_TYPE(  # type: ignore
                TransportAddressIPv6(snmpTargetAddrTAddress)
            ).set_local_address(TransportAddressIPv6(snmpSourceAddrTAddress))

        nameToTargetMap[snmpTargetAddrName] = (
            snmpTargetAddrTDomain,
            addr,
            snmpTargetAddrTimeout,
            snmpTargetAddrRetryCount,
            snmpTargetAddrParams,
        )

        cache["id"] = snmpTargetAddrEntry.branchVersionId

    return nameToTargetMap[snmpTargetAddrName]


def get_target_parameters(snmpEngine: SnmpEngine, paramsName):
    """Return security parameters for a given target."""
    mibBuilder = snmpEngine.get_mib_builder()

    (snmpTargetParamsEntry,) = mibBuilder.import_symbols(  # type: ignore
        "SNMP-TARGET-MIB", "snmpTargetParamsEntry"
    )

    cache: "dict[str, Any] | None" = snmpEngine.get_user_context("getTargetParams")
    if cache is None:
        cache = {"id": -1}
        snmpEngine.set_user_context(getTargetParams=cache)

    if cache["id"] != snmpTargetParamsEntry.branchVersionId:
        cache["nameToParamsMap"] = {}

    nameToParamsMap = cache["nameToParamsMap"]

    if paramsName not in nameToParamsMap:
        (
            snmpTargetParamsMPModel,
            snmpTargetParamsSecurityModel,
            snmpTargetParamsSecurityName,
            snmpTargetParamsSecurityLevel,
        ) = mibBuilder.import_symbols(  # type: ignore
            "SNMP-TARGET-MIB",
            "snmpTargetParamsMPModel",
            "snmpTargetParamsSecurityModel",
            "snmpTargetParamsSecurityName",
            "snmpTargetParamsSecurityLevel",
        )

        tblIdx = snmpTargetParamsEntry.getInstIdFromIndices(paramsName)

        try:
            snmpTargetParamsMPModel = snmpTargetParamsMPModel.getNode(
                snmpTargetParamsMPModel.name + tblIdx
            ).syntax
            snmpTargetParamsSecurityModel = snmpTargetParamsSecurityModel.getNode(
                snmpTargetParamsSecurityModel.name + tblIdx
            ).syntax
            snmpTargetParamsSecurityName = snmpTargetParamsSecurityName.getNode(
                snmpTargetParamsSecurityName.name + tblIdx
            ).syntax
            snmpTargetParamsSecurityLevel = snmpTargetParamsSecurityLevel.getNode(
                snmpTargetParamsSecurityLevel.name + tblIdx
            ).syntax
        except NoSuchInstanceError:
            raise SmiError("Parameters %s not configured at LCD" % paramsName)

        nameToParamsMap[paramsName] = (
            snmpTargetParamsMPModel,
            snmpTargetParamsSecurityModel,
            snmpTargetParamsSecurityName,
            snmpTargetParamsSecurityLevel,
        )

        cache["id"] = snmpTargetParamsEntry.branchVersionId

    return nameToParamsMap[paramsName]


def get_target_info(snmpEngine: SnmpEngine, snmpTargetAddrName):
    """Return transport endpoint and security parameters for a given target."""
    # Transport endpoint
    (
        snmpTargetAddrTDomain,
        snmpTargetAddrTAddress,
        snmpTargetAddrTimeout,
        snmpTargetAddrRetryCount,
        snmpTargetAddrParams,
    ) = get_target_address(snmpEngine, snmpTargetAddrName)

    (
        snmpTargetParamsMPModel,
        snmpTargetParamsSecurityModel,
        snmpTargetParamsSecurityName,
        snmpTargetParamsSecurityLevel,
    ) = get_target_parameters(snmpEngine, snmpTargetAddrParams)

    return (
        snmpTargetAddrTDomain,
        snmpTargetAddrTAddress,
        snmpTargetAddrTimeout,
        snmpTargetAddrRetryCount,
        snmpTargetParamsMPModel,
        snmpTargetParamsSecurityModel,
        snmpTargetParamsSecurityName,
        snmpTargetParamsSecurityLevel,
    )


def get_notification_info(snmpEngine: SnmpEngine, notificationTarget):
    """Return notification tag and type for a given target."""
    mibBuilder = snmpEngine.get_mib_builder()

    (snmpNotifyEntry,) = mibBuilder.import_symbols(  # type: ignore
        "SNMP-NOTIFICATION-MIB", "snmpNotifyEntry"
    )

    cache: "dict[str, Any] | None" = snmpEngine.get_user_context("getNotificationInfo")
    if cache is None:
        cache = {"id": -1}
        snmpEngine.set_user_context(getNotificationInfo=cache)

    if cache["id"] != snmpNotifyEntry.branchVersionId:
        cache["targetToNotifyMap"] = {}  # type: ignore

    targetToNotifyMap = cache["targetToNotifyMap"]

    if notificationTarget not in targetToNotifyMap:
        (snmpNotifyTag, snmpNotifyType) = mibBuilder.import_symbols(  # type: ignore
            "SNMP-NOTIFICATION-MIB", "snmpNotifyTag", "snmpNotifyType"
        )

        tblIdx = snmpNotifyEntry.getInstIdFromIndices(notificationTarget)

        try:
            snmpNotifyTag = snmpNotifyTag.getNode(snmpNotifyTag.name + tblIdx).syntax
            snmpNotifyType = snmpNotifyType.getNode(snmpNotifyType.name + tblIdx).syntax

        except NoSuchInstanceError:
            raise SmiError("Target %s not configured at LCD" % notificationTarget)

        targetToNotifyMap[notificationTarget] = (snmpNotifyTag, snmpNotifyType)

        cache["id"] = snmpNotifyEntry.branchVersionId

    return targetToNotifyMap[notificationTarget]


def get_target_names(snmpEngine: SnmpEngine, tag):
    """Return a list of target names associated with a given tag."""
    mibBuilder = snmpEngine.get_mib_builder()

    (snmpTargetAddrEntry,) = mibBuilder.import_symbols(  # type: ignore
        "SNMP-TARGET-MIB", "snmpTargetAddrEntry"
    )

    cache: "dict[str, Any] | None" = snmpEngine.get_user_context("getTargetNames")
    if cache is None:
        cache = {"id": -1}
        snmpEngine.set_user_context(getTargetNames=cache)

    if cache["id"] == snmpTargetAddrEntry.branchVersionId:
        tagToTargetsMap = cache["tagToTargetsMap"]
    else:
        cache["tagToTargetsMap"] = {}

        tagToTargetsMap = cache["tagToTargetsMap"]

        (
            SnmpTagValue,
            snmpTargetAddrName,
            snmpTargetAddrTagList,
        ) = mibBuilder.import_symbols(  # type: ignore
            "SNMP-TARGET-MIB",
            "SnmpTagValue",
            "snmpTargetAddrName",
            "snmpTargetAddrTagList",
        )
        mibNode = snmpTargetAddrTagList
        while True:
            try:
                mibNode = snmpTargetAddrTagList.getNextNode(mibNode.name)
            except NoSuchInstanceError:
                break

            idx = mibNode.name[len(snmpTargetAddrTagList.name) :]

            _snmpTargetAddrName = snmpTargetAddrName.getNode(
                snmpTargetAddrName.name + idx
            ).syntax

            for _tag in mibNode.syntax.asOctets().split():
                _tag = SnmpTagValue(_tag)
                if _tag not in tagToTargetsMap:
                    tagToTargetsMap[_tag] = []
                tagToTargetsMap[_tag].append(_snmpTargetAddrName)

        cache["id"] = snmpTargetAddrEntry.branchVersionId

    if tag not in tagToTargetsMap:
        raise SmiError("Transport tag %s not configured at LCD" % tag)

    return tagToTargetsMap[tag]


# convert cmdrsp/cmdgen into this api



"""
SNMPv1 TRAP
+++++++++++

Send SNMP notification using the following options:

* SNMPv1
* with community name 'public'
* over IPv4/UDP
* to a Manager at 127.0.0.1:162
* send TRAP notification
* with TRAP ID 'coldStart' specified as an OID
* include managed objects information:
* overriding Uptime value with 12345
* overriding Agent Address with '127.0.0.1'
* overriding Enterprise OID with 1.3.6.1.4.1.20408.4.1.1.2
* include managed object information '1.3.6.1.2.1.1.1.0' = 'my system'

Functionally similar to:

| $ snmptrap -v1 -c public demo.pysnmp.com 1.3.6.1.4.1.20408.4.1.1.2 127.0.0.1 6 432 12345 1.3.6.1.2.1.1.1.0 s "my system"

"""  #
from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import ntforg
from pysnmp.proto.api import v2c

# Create SNMP engine instance
snmpEngine = engine.SnmpEngine()

# SecurityName <-> CommunityName mapping
config.add_v1_system(snmpEngine, "my-area", "public", transportTag="all-my-managers")

# Specify security settings per SecurityName (SNMPv1 -> 0)
config.add_target_parameters(snmpEngine, "my-creds", "my-area", "noAuthNoPriv", 0)

# Setup transport endpoint and bind it with security settings yielding
# a target name
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)
config.add_target_address(
    snmpEngine,
    "my-nms",
    udp.DOMAIN_NAME,
    ("127.0.0.1", 162),
    "my-creds",
    tagList="all-my-managers",
)

# Specify what kind of notification should be sent (TRAP or INFORM),
# to what targets (chosen by tag) and what filter should apply to
# the set of targets (selected by tag)
config.add_notification_target(
    snmpEngine, "my-notification", "my-filter", "all-my-managers", "trap"
)

# Allow NOTIFY access to Agent's MIB by this SNMP model (1), securityLevel
# and SecurityName
config.add_context(snmpEngine, "")
config.add_vacm_user(snmpEngine, 1, "my-area", "noAuthNoPriv", (), (), (1, 3, 6))

# *** SNMP engine configuration is complete by this line ***

# Create Notification Originator App instance.
ntfOrg = ntforg.NotificationOriginator()

# Build and submit notification message to dispatcher
ntfOrg.send_varbinds(
    snmpEngine,
    # Notification targets
    "my-notification",  # notification targets
    None,
    "",  # contextEngineId, contextName
    # var-binds
    [
        # Uptime value with 12345
        (v2c.ObjectIdentifier("1.3.6.1.2.1.1.3.0"), v2c.TimeTicks(12345)),
        # trap OID: Generic Trap #6 (enterpriseSpecific)
        #           and Specific Trap 432
        (
            v2c.ObjectIdentifier("1.3.6.1.6.3.1.1.4.1.0"),
            v2c.ObjectIdentifier("1.3.6.1.4.1.20408.4.1.1.2.0.432"),
        ),
        # Agent Address with '127.0.0.1'
        (v2c.ObjectIdentifier("1.3.6.1.6.3.18.1.3.0"), v2c.IpAddress("127.0.0.1")),
        # Enterprise OID with 1.3.6.1.4.1.20408.4.1.1.2
        (
            v2c.ObjectIdentifier("1.3.6.1.6.3.1.1.4.3.0"),
            v2c.ObjectIdentifier("1.3.6.1.4.1.20408.4.1.1.2"),
        ),
        # managed object '1.3.6.1.2.1.1.1.0' = 'my system'
        (v2c.ObjectIdentifier("1.3.6.1.2.1.1.1.0"), v2c.OctetString("my system")),
    ],
)

print("Notification is scheduled to be sent")

# Run I/O dispatcher which would send pending message and stop
snmpEngine.open_dispatcher()



"""
Agent operations on MIB
+++++++++++++++++++++++

This script explains how SNMP Agent application manipulates
its MIB possibly triggered by SNMP Manager's commands.

"""  #
# SNMP agent backend e.g. Agent access to Managed Objects
from pysnmp.smi import builder, instrum, exval

print("Loading MIB modules..."),
mibBuilder = builder.MibBuilder().load_modules(
    "SNMPv2-MIB", "SNMP-FRAMEWORK-MIB", "SNMP-COMMUNITY-MIB"
)
print("done")

print("Building MIB tree..."),
mibInstrum = instrum.MibInstrumController(mibBuilder)
print("done")

print("Building table entry index from human-friendly representation..."),
(snmpCommunityEntry,) = mibBuilder.import_symbols(
    "SNMP-COMMUNITY-MIB", "snmpCommunityEntry"
)
instanceId = snmpCommunityEntry.getInstIdFromIndices("my-router")
print("done")

print("Create/update SNMP-COMMUNITY-MIB::snmpCommunityEntry table row: ")
varBinds = mibInstrum.write_variables(
    (snmpCommunityEntry.name + (2,) + instanceId, "mycomm"),
    (snmpCommunityEntry.name + (3,) + instanceId, "mynmsname"),
    (snmpCommunityEntry.name + (7,) + instanceId, "volatile"),
)
for oid, val in varBinds:
    print(
        "{} = {}".format(
            ".".join([str(x) for x in oid]),
            not val.isValue and "N/A" or val.prettyPrint(),
        )
    )
print("done")

print("Read whole MIB (table walk)")
varBinds = [((), None)]
while True:
    varBinds = mibInstrum.read_next_variables(*varBinds)
    oid, val = varBinds[0]
    if exval.endOfMib.isSameTypeWith(val):
        break
    print(
        "{} = {}".format(
            ".".join([str(x) for x in oid]),
            not val.isValue and "N/A" or val.prettyPrint(),
        )
    )
print("done")

print("Unloading MIB modules..."),
mibBuilder.unload_modules()
print("done")



"""
Walk multiple Agents at once
++++++++++++++++++++++++++++

Iterate over MIBs of multiple SNMP Agents asynchronously using the
following options:

* with SNMPv1, community 'public' and
  with SNMPv2c, community 'public' and
  with SNMPv3, user 'usr-md5-des', MD5 auth and DES privacy
* over IPv4/UDP and
  over IPv6/UDP
* to an Agent at demo.snmplabs.com:161 and
  to an Agent at [::1]:161
* pull variables till EOM

"""  #
import asyncio

from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    # List of targets in the following format:
    # ((authData, transportTarget, varNames), ...)
    targets = (
        # 1-st target (SNMPv1 over IPv4/UDP)
        (
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
            ObjectType(ObjectIdentity("1.3.6.1.2.1")),
        ),
        # 2-nd target (SNMPv2c over IPv4/UDP)
        (
            CommunityData("public"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
            ObjectType(ObjectIdentity("1.3.6.1.4.1")),
        ),
        # 3-nd target (SNMPv3 over IPv4/UDP)
        (
            UsmUserData("usr-md5-des", "authkey1", "privkey1"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
            ObjectType(ObjectIdentity("SNMPv2-MIB", "system")),
        ),
        # 4-th target (SNMPv3 over IPv6/UDP)
        (
            UsmUserData("usr-md5-none", "authkey1"),
            await Udp6TransportTarget.create(("demo.pysnmp.com", 161)),
            ObjectType(ObjectIdentity("IF-MIB", "ifTable")),
        )
        # N-th target
        # ...
    )

    snmpEngine = SnmpEngine()

    # Submit initial GETNEXT requests and wait for responses
    for authData, transportTarget, varBind in targets:
        async for (errorIndication, errorStatus, errorIndex, varBindTable) in walk_cmd(
            snmpEngine, authData, transportTarget, ContextData(), varBind
        ):
            if errorIndication:
                print(errorIndication)
                return

            elif errorStatus:
                print(
                    "{} at {}".format(
                        errorStatus.prettyPrint(),
                        errorIndex and varBindTable[int(errorIndex) - 1][0] or "?",
                    )
                )
                return

            else:
                for varBind in varBindTable:
                    print(" = ".join([x.prettyPrint() for x in varBind]))

    snmpEngine.transport_dispatcher.run_dispatcher()


asyncio.run(run())



import pytest
from pysnmp.hlapi.v1arch.asyncio import *
from tests.agent_context import AGENT_PORT, AgentContextManager


total_count = 212  # 267


@pytest.mark.asyncio
async def test_v2_walk():  # some agents have different v2 GET NEXT behavior
    async with AgentContextManager():
        with SnmpDispatcher() as snmpDispatcher:
            objects = walk_cmd(
                snmpDispatcher,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            objects_list = [item async for item in objects]

            errorIndication, errorStatus, errorIndex, varBinds = objects_list[0]

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"

            errorIndication, errorStatus, errorIndex, varBinds = objects_list[1]

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysUpTime.0"

            assert len(objects_list) == total_count


@pytest.mark.asyncio
async def test_v2_walk_subtree():
    async with AgentContextManager():
        with SnmpDispatcher() as snmpDispatcher:
            objects = walk_cmd(
                snmpDispatcher,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "system")),
                lexicographicMode=False,
            )

            objects_list = [item async for item in objects]

            errorIndication, errorStatus, errorIndex, varBinds = objects_list[0]

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"

            errorIndication, errorStatus, errorIndex, varBinds = objects_list[1]

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"

            assert len(objects_list) == 8



import pytest
from pyasn1.type.namedtype import NamedType

from pysnmp.proto.rfc1155 import NetworkAddress, IpAddress, TypeCoercionHackMixIn


def test_clone_none():
    n = NetworkAddress()

    assert n.clone().prettyPrint() == n.prettyPrint()


def test_clone_NetworkAddress():
    n = NetworkAddress()

    assert n.clone(n.clone("10.10.10.10")).getName() == "internet"


def test_clone_IpAddress():
    ip = IpAddress("10.10.10.10")
    n = NetworkAddress()

    assert n.clone(ip).getName() == "internet"


def test_clone_string():
    n = NetworkAddress()

    assert n.clone("10.10.10.10").getName() == "internet"


def test_verifyComponent_normal():
    t = TypeCoercionHackMixIn()
    t._componentType = [NamedType("internet", IpAddress("10.10.10.10"))]

    t._verify_component(0, IpAddress("10.2.3.4"))


def test_verifyComponent_invalidIdx():
    t = TypeCoercionHackMixIn()
    t._componentType = [NamedType("internet", IpAddress("10.10.10.10"))]

    with pytest.raises(Exception):
        t._verify_component(1, IpAddress("10.2.3.4"))



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.type import namedval

mibBuilder.export_symbols("ASN1-ENUMERATION", NamedValues=namedval.NamedValues)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module PYSNMP-USM-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source http://mibs.pysnmp.com:80/asn1/PYSNMP-USM-MIB
# Produced by pysmi-0.1.3 at Mon Apr 17 11:46:02 2017
# On host grommit.local platform Darwin version 16.4.0 by user ilya
# Using Python version 3.4.2 (v3.4.2:ab2c023a9432, Oct  5 2014, 20:42:22)
#
Integer, OctetString, ObjectIdentifier = mibBuilder.import_symbols(
    "ASN1", "Integer", "OctetString", "ObjectIdentifier"
)
(NamedValues,) = mibBuilder.import_symbols("ASN1-ENUMERATION", "NamedValues")
(
    SingleValueConstraint,
    ValueRangeConstraint,
    ConstraintsIntersection,
    ValueSizeConstraint,
    ConstraintsUnion,
) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ConstraintsIntersection",
    "ValueSizeConstraint",
    "ConstraintsUnion",
)
(pysnmpModuleIDs,) = mibBuilder.import_symbols("PYSNMP-MIB", "pysnmpModuleIDs")
(SnmpAdminString,) = mibBuilder.import_symbols("SNMP-FRAMEWORK-MIB", "SnmpAdminString")
(usmUserEntry,) = mibBuilder.import_symbols("SNMP-USER-BASED-SM-MIB", "usmUserEntry")
NotificationGroup, ModuleCompliance = mibBuilder.import_symbols(
    "SNMPv2-CONF", "NotificationGroup", "ModuleCompliance"
)
(
    ModuleIdentity,
    iso,
    MibScalar,
    MibTable,
    MibTableRow,
    MibTableColumn,
    Gauge32,
    NotificationType,
    IpAddress,
    MibIdentifier,
    Unsigned32,
    Counter32,
    ObjectIdentity,
    Counter64,
    Bits,
    Integer32,
    TimeTicks,
) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "ModuleIdentity",
    "iso",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "Gauge32",
    "NotificationType",
    "IpAddress",
    "MibIdentifier",
    "Unsigned32",
    "Counter32",
    "ObjectIdentity",
    "Counter64",
    "Bits",
    "Integer32",
    "TimeTicks",
)
RowStatus, DisplayString, TextualConvention = mibBuilder.import_symbols(
    "SNMPv2-TC", "RowStatus", "DisplayString", "TextualConvention"
)
pysnmpUsmMIB = ModuleIdentity((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1))
if mibBuilder.loadTexts:
    pysnmpUsmMIB.setRevisions(
        (
            "2017-04-14 00:00",
            "2005-05-14 00:00",
        )
    )
if mibBuilder.loadTexts:
    pysnmpUsmMIB.setLastUpdated("201704140000Z")
if mibBuilder.loadTexts:
    pysnmpUsmMIB.setOrganization("The PySNMP Project")
if mibBuilder.loadTexts:
    pysnmpUsmMIB.setContactInfo(
        "E-mail: LeXtudio Inc. <support@lextudio.com> GitHub: https://github.com/lextudio/pysnmp"
    )
if mibBuilder.loadTexts:
    pysnmpUsmMIB.setDescription(
        "This MIB module defines objects specific to User Security Model (USM) implementation at PySNMP."
    )
pysnmpUsmMIBObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1))
pysnmpUsmMIBConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 2))
pysnmpUsmCfg = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 1))
pysnmpUsmDiscoverable = MibScalar(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 1, 1),
    Integer32()
    .subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1)))
    .clone(namedValues=NamedValues(("notDiscoverable", 0), ("discoverable", 1)))
    .clone("discoverable"),
).setMaxAccess(")")
if mibBuilder.loadTexts:
    pysnmpUsmDiscoverable.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmDiscoverable.setDescription(
        "Whether SNMP engine would support its discovery by responding to unknown clients."
    )
pysnmpUsmDiscovery = MibScalar(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 1, 2),
    Integer32()
    .subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1)))
    .clone(namedValues=NamedValues(("doNotDiscover", 0), ("doDiscover", 1)))
    .clone("doDiscover"),
).setMaxAccess("read-write")
if mibBuilder.loadTexts:
    pysnmpUsmDiscovery.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmDiscovery.setDescription(
        "Whether SNMP engine would try to figure out the EngineIDs of its peers by sending discover requests."
    )
pysnmpUsmKeyType = MibScalar(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 1, 3),
    Integer32()
    .subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(0, 1, 2)))
    .clone(namedValues=NamedValues(("passphrase", 0), ("master", 1), ("localized", 2)))
    .clone("passphrase"),
).setMaxAccess("not-accessible")
if mibBuilder.loadTexts:
    pysnmpUsmKeyType.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyType.setDescription(
        'When configuring USM user, the value of this enumeration determines how the keys should be treated. The default value "passphrase" means that given keys are plain-text pass-phrases, "master" indicates that the keys are pre-hashed pass-phrases, while "localized" stands for pre-hashed pass-phrases mixed with SNMP Security Engine ID value.'
    )
pysnmpUsmUser = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3))
pysnmpUsmSecretTable = MibTable(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2),
)
if mibBuilder.loadTexts:
    pysnmpUsmSecretTable.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretTable.setDescription(
        "The table of USM users passphrases configured in the SNMP engine's Local Configuration Datastore (LCD)."
    )
pysnmpUsmSecretEntry = MibTableRow(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2, 1),
).setIndexNames((1, "PYSNMP-USM-MIB", "pysnmpUsmSecretUserName"))
if mibBuilder.loadTexts:
    pysnmpUsmSecretEntry.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretEntry.setDescription(
        "Information about a particular USM user credentials."
    )
pysnmpUsmSecretUserName = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2, 1, 1),
    SnmpAdminString().subtype(subtypeSpec=ValueSizeConstraint(1, 32)),
)
if mibBuilder.loadTexts:
    pysnmpUsmSecretUserName.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretUserName.setDescription(
        "The username string for which a row in this table represents a configuration."
    )
pysnmpUsmSecretAuthKey = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2, 1, 2),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 65535)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmSecretAuthKey.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretAuthKey.setDescription(
        "User's authentication passphrase used for localized key generation."
    )
pysnmpUsmSecretPrivKey = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2, 1, 3),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 65535)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmSecretPrivKey.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretPrivKey.setDescription(
        "User's encryption passphrase used for localized key generation."
    )
pysnmpUsmSecretStatus = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 2, 1, 4), RowStatus()
).setMaxAccess("read-create")
if mibBuilder.loadTexts:
    pysnmpUsmSecretStatus.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmSecretStatus.setDescription("Table status")
pysnmpUsmKeyTable = MibTable(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3),
)
if mibBuilder.loadTexts:
    pysnmpUsmKeyTable.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyTable.setDescription(
        "The table of USM users localized keys configured in the SNMP engine's Local Configuration Datastore (LCD)."
    )
pysnmpUsmKeyEntry = MibTableRow(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3, 1),
)
usmUserEntry.registerAugmentions(("PYSNMP-USM-MIB", "pysnmpUsmKeyEntry"))
pysnmpUsmKeyEntry.setIndexNames(*usmUserEntry.getIndexNames())
if mibBuilder.loadTexts:
    pysnmpUsmKeyEntry.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyEntry.setDescription(
        "Information about a particular USM user credentials."
    )
pysnmpUsmKeyAuthLocalized = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3, 1, 1),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 64)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmKeyAuthLocalized.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyAuthLocalized.setDescription(
        "User's localized key used for authentication."
    )
pysnmpUsmKeyPrivLocalized = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3, 1, 2),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 64)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmKeyPrivLocalized.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyPrivLocalized.setDescription(
        "User's localized key used for encryption."
    )
pysnmpUsmKeyAuth = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3, 1, 3),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 64)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmKeyAuth.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyAuth.setDescription("User's non-localized key used for authentication.")
pysnmpUsmKeyPriv = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 1, 3, 1, 4),
    OctetString("\x00\x00\x00\x00\x00\x00\x00\x00").subtype(
        subtypeSpec=ValueSizeConstraint(8, 64)
    ),
)
if mibBuilder.loadTexts:
    pysnmpUsmKeyPriv.setStatus("current")
if mibBuilder.loadTexts:
    pysnmpUsmKeyPriv.setDescription("User's non-localized key used for encryption.")
pysnmpUsmMIBCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 2, 1))
pysnmpUsmMIBGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 1, 2, 2))
mibBuilder.export_symbols(
    "PYSNMP-USM-MIB",
    pysnmpUsmCfg=pysnmpUsmCfg,
    pysnmpUsmDiscoverable=pysnmpUsmDiscoverable,
    pysnmpUsmKeyType=pysnmpUsmKeyType,
    pysnmpUsmKeyEntry=pysnmpUsmKeyEntry,
    pysnmpUsmKeyTable=pysnmpUsmKeyTable,
    pysnmpUsmKeyPrivLocalized=pysnmpUsmKeyPrivLocalized,
    pysnmpUsmMIBCompliances=pysnmpUsmMIBCompliances,
    pysnmpUsmMIBObjects=pysnmpUsmMIBObjects,
    pysnmpUsmSecretTable=pysnmpUsmSecretTable,
    PYSNMP_MODULE_ID=pysnmpUsmMIB,
    pysnmpUsmSecretEntry=pysnmpUsmSecretEntry,
    pysnmpUsmMIBConformance=pysnmpUsmMIBConformance,
    pysnmpUsmUser=pysnmpUsmUser,
    pysnmpUsmKeyAuth=pysnmpUsmKeyAuth,
    pysnmpUsmSecretPrivKey=pysnmpUsmSecretPrivKey,
    pysnmpUsmKeyAuthLocalized=pysnmpUsmKeyAuthLocalized,
    pysnmpUsmMIB=pysnmpUsmMIB,
    pysnmpUsmDiscovery=pysnmpUsmDiscovery,
    pysnmpUsmSecretUserName=pysnmpUsmSecretUserName,
    pysnmpUsmKeyPriv=pysnmpUsmKeyPriv,
    pysnmpUsmSecretAuthKey=pysnmpUsmSecretAuthKey,
    pysnmpUsmSecretStatus=pysnmpUsmSecretStatus,
    pysnmpUsmMIBGroups=pysnmpUsmMIBGroups,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from hashlib import md5, sha1

from pyasn1.type import univ


def hash_passphrase(passphrase, hashFunc) -> univ.OctetString:
    """Return hash of passphrase using hashFunc hash function."""
    passphrase = univ.OctetString(passphrase).asOctets()
    # noinspection PyDeprecation,PyCallingNonCallable
    hasher = hashFunc()
    ringBuffer = passphrase * (64 // len(passphrase) + 1)
    # noinspection PyTypeChecker
    ringBufferLen = len(ringBuffer)
    count = 0
    mark = 0
    while count < 16384:
        e = mark + 64
        if e < ringBufferLen:
            hasher.update(ringBuffer[mark:e])
            mark = e
        else:
            hasher.update(
                ringBuffer[mark:ringBufferLen] + ringBuffer[0 : e - ringBufferLen]
            )
            mark = e - ringBufferLen
        count += 1
    digest = hasher.digest()
    return univ.OctetString(digest)


def password_to_key(passphrase, snmpEngineId, hashFunc) -> univ.OctetString:
    """Return key from password."""
    return localize_key(hash_passphrase(passphrase, hashFunc), snmpEngineId, hashFunc)


def localize_key(passKey, snmpEngineId, hashFunc) -> univ.OctetString:
    """Localize passKey with snmpEngineId using hashFunc hash function."""
    passKey = univ.OctetString(passKey).asOctets()
    # noinspection PyDeprecation,PyCallingNonCallable
    digest = hashFunc(passKey + snmpEngineId.asOctets() + passKey).digest()
    return univ.OctetString(digest)


# RFC3414: A.2.1
def hash_passphrase_md5(passphrase) -> univ.OctetString:
    """Return MD5 hash of passphrase."""
    return hash_passphrase(passphrase, md5)


# RFC3414: A.2.2
def hash_passphrase_sha(passphrase) -> univ.OctetString:
    """Return SHA-1 hash of passphrase."""
    return hash_passphrase(passphrase, sha1)


def password_to_key_md5(passphrase, snmpEngineId) -> univ.OctetString:
    """Return MD5 key from password."""
    return localize_key(hash_passphrase_md5(passphrase), snmpEngineId, md5)


def password_to_key_sha(passphrase, snmpEngineId) -> univ.OctetString:
    """Return SHA-1 key from password."""
    return localize_key(hash_passphrase_sha(passphrase), snmpEngineId, sha1)


def localize_key_md5(passKey, snmpEngineId) -> univ.OctetString:
    """Localize passKey with snmpEngineId using MD5 hash function."""
    return localize_key(passKey, snmpEngineId, md5)


def localize_key_sha(passKey, snmpEngineId) -> univ.OctetString:
    """Localize passKey with snmpEngineId using SHA1 hash function."""
    return localize_key(passKey, snmpEngineId, sha1)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMP-USER-BASED-SM-3DES-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/SNMP-USER-BASED-SM-3DES-MIB
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:36 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(snmpPrivProtocols,) = mibBuilder.import_symbols(
    "SNMP-FRAMEWORK-MIB",
    "snmpPrivProtocols")

(ModuleCompliance,
 NotificationGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 snmpModules) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "snmpModules")

(AutonomousType,
 DisplayString,
 TextualConvention) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "AutonomousType",
    "DisplayString",
    "TextualConvention")


# MODULE-IDENTITY

snmpUsmMIB = ModuleIdentity(
    (1, 3, 6, 1, 6, 3, 15)
)
if mibBuilder.loadTexts:
    snmpUsmMIB.setRevisions(
        ("1999-10-06 00:00",)
    )
if mibBuilder.loadTexts:
    snmpUsmMIB.setLastUpdated("9910060000Z")
if mibBuilder.loadTexts:
    snmpUsmMIB.setOrganization("SNMPv3 Working Group")
if mibBuilder.loadTexts:
    snmpUsmMIB.setContactInfo("WG-email: snmpv3@lists.tislabs.com Subscribe: majordomo@lists.tislabs.com In msg body: subscribe snmpv3 Chair: Russ Mundy NAI Labs postal: 3060 Washington Rd Glenwood MD 21738 USA email: mundy@tislabs.com phone: +1-443-259-2307 Co-editor: David Reeder NAI Labs postal: 3060 Washington Road (Route 97) Glenwood, MD 21738 USA email: dreeder@tislabs.com phone: +1-443-259-2348 Co-editor: Olafur Gudmundsson NAI Labs postal: 3060 Washington Road (Route 97) Glenwood, MD 21738 USA email: ogud@tislabs.com phone: +1-443-259-2389 ")
if mibBuilder.loadTexts:
    snmpUsmMIB.setDescription("Extension to the SNMP User-based Security Model to support Triple-DES EDE in 'Outside' CBC (cipher-block chaining) Mode. ")


# Types definitions


# TEXTUAL-CONVENTIONS



# MIB Managed Objects in the order of their OIDs

_Usm3DESEDEPrivProtocol_ObjectIdentity = ObjectIdentity
usm3DESEDEPrivProtocol = _Usm3DESEDEPrivProtocol_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 10, 1, 2, 3)
)
if mibBuilder.loadTexts:
    usm3DESEDEPrivProtocol.setStatus("current")
if mibBuilder.loadTexts:
    usm3DESEDEPrivProtocol.setReference("- Data Encryption Standard, National Institute of Standards and Technology. Federal Information Processing Standard (FIPS) Publication 46-3, (1999, pending approval). Will supersede FIPS Publication 46-2. - Data Encryption Algorithm, American National Standards Institute. ANSI X3.92-1981, (December, 1980). - DES Modes of Operation, National Institute of Standards and Technology. Federal Information Processing Standard (FIPS) Publication 81, (December, 1980). - Data Encryption Algorithm - Modes of Operation, American National Standards Institute. ANSI X3.106-1983, (May 1983). ")
if mibBuilder.loadTexts:
    usm3DESEDEPrivProtocol.setDescription("The 3DES-EDE Symmetric Encryption Protocol.")

# Managed Objects groups


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMP-USER-BASED-SM-3DES-MIB",
    **{"usm3DESEDEPrivProtocol": usm3DESEDEPrivProtocol,
       "snmpUsmMIB": snmpUsmMIB,
       "PYSNMP_MODULE_ID": snmpUsmMIB}
)



"""
Walk whole MIB
++++++++++++++

Send a series of SNMP GETNEXT requests using the following options:

* with SNMPv3, user 'usr-md5-none', MD5 authentication, no privacy
* over IPv4/UDP
* to an Agent at demo.pysnmp.com:161
* for all OIDs in IF-MIB

Functionally similar to:

| $ snmpwalk -v3 -lauthNoPriv -u usr-md5-none -A authkey1 -X privkey1 \
|            demo.pysnmp.com  IF-MIB::

"""  #
import asyncio

from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    snmpEngine = SnmpEngine()

    async for (errorIndication, errorStatus, errorIndex, varBindTable) in walk_cmd(
        snmpEngine,
        UsmUserData("usr-md5-none", "authkey1"),
        await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
        ContextData(),
        ObjectType(ObjectIdentity("IF-MIB", "ifTable")),
    ):
        if errorIndication:
            print(errorIndication)
            return

        elif errorStatus:
            print(
                "{} at {}".format(
                    errorStatus.prettyPrint(),
                    errorIndex and varBindTable[int(errorIndex) - 1][0] or "?",
                )
            )
            return

        else:
            for varBind in varBindTable:
                print(" = ".join([x.prettyPrint() for x in varBind]))

    snmpEngine.transport_dispatcher.run_dispatcher()


asyncio.run(run())
